#!/usr/bin/env python
# Copyright (c) 2015-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Reader loading."""
import logging

import yaml

from satpy.readers.core.yaml_reader import AbstractYAMLReader

from .config import configs_for_reader

LOG = logging.getLogger(__name__)


def load_readers(filenames=None, reader=None, reader_kwargs=None):
    """Create specified readers and assign files to them.

    Args:
        filenames (Iterable or dict): A sequence of files that will be used to load data from. A ``dict`` object
                                      should map reader names to a list of filenames for that reader.
        reader (str or list): The name of the reader to use for loading the data or a list of names.
        reader_kwargs (dict): Keyword arguments to pass to specific reader instances.
            This can either be a single dictionary that will be passed to all
            reader instances, or a mapping of reader names to dictionaries.  If
            the keys of ``reader_kwargs`` match exactly the list of strings in
            ``reader`` or the keys of filenames, each reader instance will get its
            own keyword arguments accordingly.

    Returns: Dictionary mapping reader name to reader instance

    """
    reader_instances = {}
    if _early_exit(filenames, reader):
        return {}

    reader, filenames, remaining_filenames = _get_reader_and_filenames(reader, filenames)
    (reader_kwargs, reader_kwargs_without_filter) = _get_reader_kwargs(reader, reader_kwargs)

    if reader_kwargs is None:
        reader_kwargs = {}

    for idx, reader_configs in enumerate(configs_for_reader(reader)):
        readers_files = _get_readers_files(filenames, reader, idx, remaining_filenames)
        reader_instance = _get_reader_instance(reader, reader_configs, idx, reader_kwargs)
        if reader_instance is None or not readers_files:
            # Reader initiliasation failed or no files were given
            continue

        loadables = reader_instance.select_files_from_pathnames(readers_files)
        if loadables:
            reader_instance.create_storage_items(
                    loadables,
                    fh_kwargs=reader_kwargs_without_filter[None if reader is None else reader[idx]])
            reader_instances[reader_instance.name] = reader_instance
            remaining_filenames -= set(loadables)

        if not remaining_filenames:
            break

    _check_remaining_files(remaining_filenames)
    _check_reader_instances(reader_instances)
    return reader_instances


def _get_readers_files(filenames, reader, idx, remaining_filenames):
    if isinstance(filenames, dict):
        return set(filenames[reader[idx]])
    return remaining_filenames


def _get_reader_instance(reader, reader_configs, idx, reader_kwargs):
    reader_instance = None
    try:
        reader_instance = load_reader(
            reader_configs,
            **reader_kwargs[None if reader is None else reader[idx]])
    except (KeyError, IOError) as err:
        LOG.info("Cannot use %s", str(reader_configs))
        LOG.debug(str(err))
    except yaml.constructor.ConstructorError as err:
        _log_yaml_error(reader_configs, err)

    return reader_instance


def load_reader(reader_configs, **reader_kwargs):
    """Import and setup the reader from *reader_info*."""
    return AbstractYAMLReader.from_config_files(*reader_configs, **reader_kwargs)


def _log_yaml_error(reader_configs, err):
    LOG.error("Problem with %s", str(reader_configs))
    LOG.error(str(err))


def _early_exit(filenames, reader):
    if not filenames and not reader:
        # used for an empty Scene
        return True
    _check_reader_and_filenames(reader, filenames)
    if not filenames:
        LOG.warning("'filenames' required to create readers and load data")
        return True
    return False


def _check_reader_and_filenames(reader, filenames):
    if reader and filenames is not None and not filenames:
        # user made a mistake in their glob pattern
        raise ValueError("'filenames' was provided but is empty.")


def _get_reader_and_filenames(reader, filenames):
    if reader is None and isinstance(filenames, dict):
        # filenames is a dictionary of reader_name -> filenames
        reader = list(filenames.keys())
        remaining_filenames = set(f for fl in filenames.values() for f in fl)
    elif reader and isinstance(filenames, dict):
        # filenames is a dictionary of reader_name -> filenames
        # but they only want one of the readers
        filenames = filenames[reader]
        remaining_filenames = set(filenames or [])
    else:
        remaining_filenames = set(filenames or [])
    return reader, filenames, remaining_filenames


def _check_remaining_files(remaining_filenames):
    if remaining_filenames:
        LOG.warning("Don't know how to open the following files: {}".format(str(remaining_filenames)))


def _check_reader_instances(reader_instances):
    if not reader_instances:
        raise ValueError("No supported files found")
    if not any(list(r.available_dataset_ids) for r in reader_instances.values()):
        raise ValueError("No dataset could be loaded. Either missing "
                         "requirements (such as Epilog, Prolog) or none of the "
                         "provided files match the filter parameters.")


def _get_reader_kwargs(reader, reader_kwargs):
    """Help load_readers to form reader_kwargs.

    Helper for load_readers to get reader_kwargs and
    reader_kwargs_without_filter in the desirable form.
    """
    reader_kwargs = reader_kwargs or {}

    # ensure one reader_kwargs per reader, None if not provided
    if reader is None:
        reader_kwargs = {None: reader_kwargs}
    elif reader_kwargs.keys() != set(reader):
        reader_kwargs = dict.fromkeys(reader, reader_kwargs)

    reader_kwargs_without_filter = {}
    for (k, v) in reader_kwargs.items():
        reader_kwargs_without_filter[k] = v.copy()
        reader_kwargs_without_filter[k].pop("filter_parameters", None)

    return (reader_kwargs, reader_kwargs_without_filter)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for IASI L2 reader."""

import datetime as dt
import math
import os

import numpy as np
import pytest
import xarray as xr

from satpy.readers.iasi_l2 import IASIL2HDF5

SCAN_WIDTH = 120
NUM_LEVELS = 138
NUM_SCANLINES = 10
FNAME = "W_XX-EUMETSAT-kan,iasi,metopb+kan_C_EUMS_20170920103559_IASI_PW3_02_M01_20170920102217Z_20170920102912Z.hdf"
# Structure for the test data, to be written to HDF5 file
TEST_DATA = {
    "INFO": {
        "OmC": {"data": np.zeros((NUM_SCANLINES, SCAN_WIDTH), dtype=np.float32),
                "attrs": {"long_name": "Cloud signal. Predicted average window channel 'Obs minus Calc",
                          "units": "K"}},
        "FLG_AMSUBAD": {"data": np.zeros((NUM_SCANLINES, 30), dtype=np.uint8),
                        "attrs": {}},
        "FLG_IASIBAD": {"data": np.zeros((NUM_SCANLINES, SCAN_WIDTH), dtype=np.uint8),
                        "attrs": {}},
        "FLG_MHSBAD": {"data": np.zeros((NUM_SCANLINES, SCAN_WIDTH), dtype=np.uint8),
                        "attrs": {}},
        # Not implemented in the reader
        "mdist": {"data": np.zeros((NUM_SCANLINES, SCAN_WIDTH), dtype=np.float32),
                  "attrs": {}},
    },
    "L1C": {
        "Latitude": {"data": np.zeros((NUM_SCANLINES, SCAN_WIDTH), dtype=np.float32),
                     "attrs": {"units": "degrees_north"}},
        "Longitude": {"data": np.zeros((NUM_SCANLINES, SCAN_WIDTH), dtype=np.float32),
                      "attrs": {"units": "degrees_north"}},
        "SatAzimuth": {"data": np.zeros((NUM_SCANLINES, SCAN_WIDTH), dtype=np.float32),
                       "attrs": {"units": "degrees"}},
        "SatZenith": {"data": np.zeros((NUM_SCANLINES, SCAN_WIDTH), dtype=np.float32),
                      "attrs": {"units": "degrees"}},
        "SensingTime_day": {"data": 6472 * np.ones(NUM_SCANLINES, dtype=np.uint16),
                            "attrs": {}},
        "SensingTime_msec": {"data": np.arange(37337532, 37338532, 100, dtype=np.uint32),
                             "attrs": {}},
        "SunAzimuth": {"data": np.zeros((NUM_SCANLINES, SCAN_WIDTH), dtype=np.float32),
                       "attrs": {"units": "degrees"}},
        "SunZenith": {"data": np.zeros((NUM_SCANLINES, SCAN_WIDTH), dtype=np.float32),
                      "attrs": {"units": "degrees"}},
    },
    "Maps": {
        "Height": {"data": np.zeros((NUM_SCANLINES, SCAN_WIDTH), dtype=np.float32),
                   "attrs": {"units": "m"}},
        "HeightStd": {"data": np.zeros((NUM_SCANLINES, SCAN_WIDTH), dtype=np.float32),
                      "attrs": {"units": "m"}},
    },
    "PWLR": {
        "E": {"data": np.zeros((NUM_SCANLINES, SCAN_WIDTH, 10), dtype=np.float32),
              "attrs": {"emissivity_wavenumbers": np.array([699.3, 826.4,
                                                            925.9, 1075.2,
                                                            1204.8, 1315.7,
                                                            1724.1, 2000.0,
                                                            2325.5, 2702.7],
                                                           dtype=np.float32)}},
        "O": {"data": np.zeros((NUM_SCANLINES, SCAN_WIDTH, NUM_LEVELS), dtype=np.float32),
              "attrs": {"long_name": "Ozone mixing ratio vertical profile",
                        "units": "kg/kg"}},
        "OC": {"data": np.zeros((NUM_SCANLINES, SCAN_WIDTH), dtype=np.float32),
               "attrs": {}},
        "P": {"data": np.zeros((NUM_SCANLINES, SCAN_WIDTH, NUM_LEVELS), dtype=np.float32),
              "attrs": {"long_name": "Atmospheric pressures at which the vertical profiles are given. "
                                     "Last value is the surface pressure",
                        "units": "hpa"}},
        "QE": {"data": np.zeros((NUM_SCANLINES, SCAN_WIDTH), dtype=np.float32),
               "attrs": {}},
        "QO": {"data": np.zeros((NUM_SCANLINES, SCAN_WIDTH), dtype=np.float32),
               "attrs": {}},
        "QP": {"data": np.zeros((NUM_SCANLINES, SCAN_WIDTH), dtype=np.float32),
               "attrs": {}},
        "QT": {"data": np.zeros((NUM_SCANLINES, SCAN_WIDTH), dtype=np.float32),
               "attrs": {}},
        "QTs": {"data": np.zeros((NUM_SCANLINES, SCAN_WIDTH), dtype=np.float32),
                "attrs": {}},
        "QW": {"data": np.zeros((NUM_SCANLINES, SCAN_WIDTH), dtype=np.float32),
               "attrs": {}},
        "T": {"data": np.zeros((NUM_SCANLINES, SCAN_WIDTH, NUM_LEVELS), dtype=np.float32),
              "attrs": {"long_name": "Temperature vertical profile", "units": "K"}},
        "Ts": {"data": np.zeros((NUM_SCANLINES, SCAN_WIDTH), dtype=np.float32),
               "attrs": {"long_name": "Surface skin temperature", "units": "K"}},
        "W": {"data": np.zeros((NUM_SCANLINES, SCAN_WIDTH, NUM_LEVELS), dtype=np.float32),
              "attrs": {"long_name": "Water vapour mixing ratio vertical profile", "units": "kg/kg"}},
        "WC": {"data": np.zeros((NUM_SCANLINES, SCAN_WIDTH), dtype=np.float32),
               "attrs": {"long_name": "Water vapour total columnar amount", "units": "mm"}},
    }
}


FNAME_INFO = {"start_time": dt.datetime(2017, 9, 20, 10, 22, 17),
              "end_time": dt.datetime(2017, 9, 20, 10, 29, 12),
              "processing_time": dt.datetime(2017, 9, 20, 10, 35, 59),
              "processing_location": "kan",
              "long_platform_id": "metopb",
              "instrument": "iasi",
              "platform_id": "M01"}

FTYPE_INFO = {"file_reader": IASIL2HDF5,
              "file_patterns": ["{fname}.hdf"],
              "file_type": "iasi_l2_hdf5"}


@pytest.fixture(scope="module")
def test_data(tmp_path_factory):
    """Save the test to the indicated directory."""
    import h5py
    test_file = os.path.join(tmp_path_factory.mktemp("data"), FNAME)
    with h5py.File(test_file, "w") as fid:
        # Create groups
        for grp in TEST_DATA:
            fid.create_group(grp)
            # Write datasets
            for dset in TEST_DATA[grp]:
                fid[grp][dset] = TEST_DATA[grp][dset]["data"]
                # Write dataset attributes
                for attr in TEST_DATA[grp][dset]["attrs"]:
                    fid[grp][dset].attrs[attr] = \
                        TEST_DATA[grp][dset]["attrs"][attr]
    return test_file


@pytest.fixture
def iasi_filehandler(test_data):
    """Create a filehandler."""
    return IASIL2HDF5(test_data, FNAME_INFO, FTYPE_INFO)


def test_scene(test_data):
    """Test scene creation."""
    from satpy import Scene
    scn = Scene(reader="iasi_l2", filenames=[test_data])
    assert scn.start_time is not None
    assert scn.end_time is not None
    assert scn.sensor_names
    assert "iasi" in scn.sensor_names


def test_scene_load_available_datasets(test_data):
    """Test that all datasets are available."""
    from satpy import Scene
    scn = Scene(reader="iasi_l2", filenames=[test_data])
    scn.load(scn.available_dataset_names())


def test_scene_load_pressure(test_data):
    """Test loading pressure data."""
    from satpy import Scene
    scn = Scene(reader="iasi_l2", filenames=[test_data])
    scn.load(["pressure"])
    pres = scn["pressure"].compute()
    check_pressure(pres, scn.attrs)


def test_scene_load_emissivity(test_data):
    """Test loading emissivity data."""
    from satpy import Scene
    scn = Scene(reader="iasi_l2", filenames=[test_data])
    scn.load(["emissivity"])
    emis = scn["emissivity"].compute()
    check_emissivity(emis)


def test_scene_load_sensing_times(test_data):
    """Test loading sensing times."""
    from satpy import Scene
    scn = Scene(reader="iasi_l2", filenames=[test_data])
    scn.load(["sensing_time"])
    times = scn["sensing_time"].compute()
    check_sensing_times(times)


def test_init(test_data, iasi_filehandler):
    """Test reader initialization."""
    assert iasi_filehandler.filename == test_data
    assert iasi_filehandler.finfo == FNAME_INFO
    assert iasi_filehandler.lons is None
    assert iasi_filehandler.lats is None
    assert iasi_filehandler.mda["platform_name"] == "Metop-B"
    assert iasi_filehandler.mda["sensor"] == "iasi"


def test_time_properties(iasi_filehandler):
    """Test time properties."""
    import datetime as dt
    assert isinstance(iasi_filehandler.start_time, dt.datetime)
    assert isinstance(iasi_filehandler.end_time, dt.datetime)


def test_get_dataset(iasi_filehandler):
    """Test get_dataset() for different datasets."""
    from satpy.tests.utils import make_dataid
    info = {"eggs": "spam"}
    key = make_dataid(name="pressure")
    data = iasi_filehandler.get_dataset(key, info).compute()
    check_pressure(data)
    assert "eggs" in data.attrs
    assert data.attrs["eggs"] == "spam"
    key = make_dataid(name="emissivity")
    data = iasi_filehandler.get_dataset(key, info).compute()
    check_emissivity(data)
    key = make_dataid(name="sensing_time")
    data = iasi_filehandler.get_dataset(key, info).compute()
    assert data.shape == (NUM_SCANLINES, SCAN_WIDTH)


def check_pressure(pres, attrs=None):
    """Test reading pressure dataset.

    Helper function.
    """
    assert np.all(pres == 0.0)
    assert pres.x.size == SCAN_WIDTH
    assert pres.y.size == NUM_SCANLINES
    assert pres.level.size == NUM_LEVELS
    if attrs:
        assert pres.attrs["start_time"] == attrs["start_time"]
        assert pres.attrs["end_time"] == attrs["end_time"]
    assert "long_name" in pres.attrs
    assert "units" in pres.attrs


def check_emissivity(emis):
    """Test reading emissivity dataset.

    Helper function.
    """
    assert np.all(emis == 0.0)
    assert emis.x.size == SCAN_WIDTH
    assert emis.y.size == NUM_SCANLINES
    assert "emissivity_wavenumbers" in emis.attrs


def check_sensing_times(times):
    """Test reading sensing times.

    Helper function.
    """
    # Times should be equal in blocks of four, but not beyond, so
    # there should be SCAN_WIDTH/4 different values
    for i in range(int(SCAN_WIDTH / 4)):
        assert np.unique(times[0, i * 4:i * 4 + 4]).size == 1
    assert np.unique(times[0, :]).size == SCAN_WIDTH / 4


@pytest.mark.parametrize(("dset", "dtype", "units"), [
    ("amsu_instrument_flags", np.uint8, None),
    ("iasi_instrument_flags", np.uint8, None),
    ("mhs_instrument_flags", np.uint8, None),
    ("observation_minus_calculation", np.float32, "K"),
    ("surface_elevation", np.float32, "m"),
    ("surface_elevation_std", np.float32, "m")
    ])
def test_get_info_and_maps(iasi_filehandler, dset, dtype, units):
    """Test datasets in INFO and Maps groups are read."""
    from satpy.tests.utils import make_dataid
    info = {"eggs": "spam"}
    key = make_dataid(name=dset)
    data = iasi_filehandler.get_dataset(key, info).compute()
    assert data.shape == (NUM_SCANLINES, SCAN_WIDTH)
    assert data.dtype == dtype
    if units:
        assert data.attrs["units"] == units
    assert data.attrs["platform_name"] == "Metop-B"


def test_read_dataset(test_data):
    """Test read_dataset() function."""
    import h5py

    from satpy.readers.iasi_l2 import read_dataset
    from satpy.tests.utils import make_dataid
    with h5py.File(test_data, "r") as fid:
        key = make_dataid(name="pressure")
        data = read_dataset(fid, key).compute()
        check_pressure(data)
        key = make_dataid(name="emissivity")
        data = read_dataset(fid, key).compute()
        check_emissivity(data)
        # This dataset doesn't have any attributes
        key = make_dataid(name="ozone_total_column")
        data = read_dataset(fid, key).compute()
        assert len(data.attrs) == 0


def test_read_geo(test_data):
    """Test read_geo() function."""
    import h5py

    from satpy.readers.iasi_l2 import read_geo
    from satpy.tests.utils import make_dataid
    with h5py.File(test_data, "r") as fid:
        key = make_dataid(name="sensing_time")
        data = read_geo(fid, key).compute()
        assert data.shape == (NUM_SCANLINES, SCAN_WIDTH)
        key = make_dataid(name="latitude")
        data = read_geo(fid, key).compute()
        assert data.shape == (NUM_SCANLINES, SCAN_WIDTH)


def test_form_datetimes():
    """Test _form_datetimes() function."""
    from satpy.readers.iasi_l2 import _form_datetimes
    days = TEST_DATA["L1C"]["SensingTime_day"]["data"]
    msecs = TEST_DATA["L1C"]["SensingTime_msec"]["data"]
    times = _form_datetimes(days, msecs)
    check_sensing_times(times)


@pytest.fixture
def fake_iasi_l2_cdr_nc_dataset():
    """Create minimally fake IASI L2 CDR NC dataset."""
    shp = (3, 4, 5)
    fv = -999
    dims = ("scan_lines", "pixels", "vertical_levels")
    coords2 = "latitude longitude"
    coords3 = "latitude longitude pressure_levels"
    lons = xr.DataArray(
            np.array([[0, 0, 0, 0], [1, 1, 1, 1],
                      [2, 2, 2, 2]], dtype="float32"),
            dims=dims[:2],
            attrs={"coordinates": coords2,
                   "standard_name": "longitude"})
    lats = xr.DataArray(
            np.array([[3, 3, 3, 3], [2, 2, 2, 2],
                      [1, 1, 1, 1]], dtype="float32"),
            dims=dims[:2],
            attrs={"coordinates": coords2,
                   "standard_name": "latitude"})
    pres = xr.DataArray(
            np.linspace(0, 1050, math.prod(shp), dtype="float32").reshape(shp),
            dims=dims,
            attrs={"coordinates": coords3})

    temps = np.linspace(100, 400, math.prod(shp), dtype="float32").reshape(shp)
    temps[0, 0, 0] = fv
    temp = xr.DataArray(
            temps, dims=dims,
            attrs={"coordinates": coords3, "_FillValue": fv, "units": "K"})

    iasibad = xr.DataArray(
            np.zeros(shp[:2], dtype="uint8"),
            dims=dims[:2],
            attrs={"coordinates": coords2,
                   "standard_name": "flag_information_IASI_L1c"})
    iasibad[0, 0] = 1

    cf = xr.DataArray(
            np.zeros(shp[:2], dtype="uint8"),
            dims=dims[:2],
            attrs={"coordinates": coords2,
                   "standard_name": "cloud_area_fraction",
                   "_FillValue": 255,
                   "valid_min": 0,
                   "valid_max": 100})

    return xr.Dataset(
            {"T": temp, "FLG_IASIBAD": iasibad, "CloudFraction": cf},
            coords={
                "longitude": lons,
                "latitude": lats,
                "pressure_levels": pres})


@pytest.fixture
def fake_iasi_l2_cdr_nc_file(fake_iasi_l2_cdr_nc_dataset, tmp_path):
    """Write a NetCDF file with minimal fake IASI L2 CDR NC data."""
    fn = ("W_XX-EUMETSAT-Darmstadt,HYPERSPECT+SOUNDING,METOPA+PW3+"
          "IASI_C_EUMP_19210624090000Z_19210623090100Z_eps_r_l2_0101.nc")
    of = tmp_path / fn
    fake_iasi_l2_cdr_nc_dataset.to_netcdf(of)
    return os.fspath(of)


def test_iasi_l2_cdr_nc(fake_iasi_l2_cdr_nc_file):
    """Test the IASI L2 CDR NC reader."""
    from satpy import Scene
    sc = Scene(filenames=[fake_iasi_l2_cdr_nc_file], reader=["iasi_l2_cdr_nc"])
    sc.load(["T", "FLG_IASIBAD", "CloudFraction"])
    assert sc["T"].dims == ("y", "x", "vertical_levels")
    assert sc["T"].shape == (3, 4, 5)
    assert sc["T"].attrs["area"].shape == (3, 4)
    (lons, lats) = sc["T"].attrs["area"].get_lonlats()
    np.testing.assert_array_equal(
            lons,
            np.array([[0, 0, 0, 0], [1, 1, 1, 1],
                      [2, 2, 2, 2]]))
    assert np.isnan(sc["T"][0, 0, 0])
    assert sc["FLG_IASIBAD"][0, 0] == 1
    assert sc["CloudFraction"].dtype == np.dtype("uint8")
    assert sc["T"].attrs["units"] == "K"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Satpy developers
#
# satpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# satpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with satpy.  If not, see <http://www.gnu.org/licenses/>.

"""The fci_cld_l2_nc reader tests package."""

import datetime as dt
import os
import unittest
import uuid
from contextlib import suppress
from unittest import mock

import numpy as np
import pytest
import xarray as xr
from netCDF4 import Dataset
from pyresample import geometry

from satpy.readers.fci_l2_nc import FciL2NCAMVFileHandler, FciL2NCFileHandler, FciL2NCSegmentFileHandler
from satpy.tests.utils import make_dataid

AREA_DEF = geometry.AreaDefinition(
    "mtg_fci_fdss_2km",
    "MTG FCI Full Disk Scanning Service area definition with 2 km resolution",
    "",
    {"h": 35786400., "lon_0": 0.0, "ellps": "WGS84", "proj": "geos", "units": "m"},
    5568,
    5568,
    (-5567999.9942, 5567999.9942, 5567999.9942, -5567999.9942)
)

SEG_AREA_DEF = geometry.AreaDefinition(
    "mtg_fci_fdss_32km",
    "MTG FCI Full Disk Scanning Service area definition with 32 km resolution",
    "",
    {"h": 35786400., "lon_0": 0.0, "ellps": "WGS84", "proj": "geos", "units": "m"},
    348,
    348,
    (-5567999.9942, 5567999.9942, 5567999.9942, -5567999.9942)
)


class TestFciL2NCFileHandler(unittest.TestCase):
    """Test the FciL2NCFileHandler reader."""

    def setUp(self):
        """Set up the test by creating a test file and opening it with the reader."""
        # Easiest way to test the reader is to create a test netCDF file on the fly
        # Create unique filenames to prevent race conditions when tests are run in parallel
        self.test_file = str(uuid.uuid4()) + ".nc"
        with Dataset(self.test_file, "w") as nc:
            # Create dimensions
            nc.createDimension("number_of_columns", 10)
            nc.createDimension("number_of_rows", 100)
            nc.createDimension("maximum_number_of_layers", 2)

            # add global attributes
            nc.data_source = "TEST_DATA_SOURCE"
            nc.platform = "TEST_PLATFORM"

            # Add datasets
            x = nc.createVariable("x", np.float32, dimensions=("number_of_columns",))
            x.standard_name = "projection_x_coordinate"
            x[:] = np.arange(10)

            y = nc.createVariable("y", np.float32, dimensions=("number_of_rows",))
            y.standard_name = "projection_y_coordinate"
            y[:] = np.arange(100)

            s = nc.createVariable("product_quality", np.int8)
            s[:] = 99.

            one_layer_dataset = nc.createVariable("test_one_layer", np.float32,
                                                  dimensions=("number_of_rows", "number_of_columns"))
            one_layer_dataset[:] = np.arange(1000.0).reshape((100, 10))
            one_layer_dataset.test_attr = "attr"
            one_layer_dataset.units = "test_units"

            two_layers_dataset = nc.createVariable("test_two_layers", np.float32,
                                                   dimensions=("maximum_number_of_layers",
                                                               "number_of_rows",
                                                               "number_of_columns"))
            two_layers_dataset[0, :, :] = np.ones((100, 10))
            two_layers_dataset[1, :, :] = 2 * np.ones((100, 10))
            two_layers_dataset.unit = "test_unit"

            mtg_geos_projection = nc.createVariable("mtg_geos_projection", int, dimensions=())
            mtg_geos_projection.longitude_of_projection_origin = 0.0
            mtg_geos_projection.semi_major_axis = 6378137.
            mtg_geos_projection.inverse_flattening = 298.257223563
            mtg_geos_projection.perspective_point_height = 35786400.

            # Add enumerated type
            enum_dict = {"False": 0, "True": 1}
            bool_type = nc.createEnumType(np.uint8,"bool_t",enum_dict)
            nc.createVariable("quality_flag", bool_type,
                              dimensions=("number_of_rows", "number_of_columns"))

        self.fh = FciL2NCFileHandler(filename=self.test_file, filename_info={}, filetype_info={})

    def tearDown(self):
        """Remove the previously created test file."""
        # First delete the file handler, forcing the file to be closed if still open
        del self.fh
        # Then we can safely remove the file from the system
        with suppress(OSError):
            os.remove(self.test_file)

    def test_all_basic(self):
        """Test all basic functionalities."""
        assert self.fh.spacecraft_name == "TEST_PLATFORM"
        assert self.fh.sensor_name == "test_data_source"
        assert self.fh.ssp_lon == 0.0

        global_attributes = self.fh._get_global_attributes()
        expected_global_attributes = {
            "filename": self.test_file,
            "spacecraft_name": "TEST_PLATFORM",
            "ssp_lon": 0.0,
            "sensor": "test_data_source",
            "platform_name": "TEST_PLATFORM"
        }
        assert global_attributes == expected_global_attributes

    @mock.patch("satpy.readers.fci_l2_nc.geometry.AreaDefinition")
    @mock.patch("satpy.readers.fci_l2_nc.make_ext")
    def test_area_definition(self, me_, gad_):
        """Test the area definition computation."""
        self.fh._compute_area_def(make_dataid(name="test_area_def", resolution=2000))

        # Asserts that the make_ext function was called with the correct arguments
        me_.assert_called_once()
        args, kwargs = me_.call_args
        np.testing.assert_allclose(args, [-0.0, -515.6620, 5672.28217, 0.0, 35786400.])

        proj_dict = {"a": 6378137.,
                     "lon_0": 0.0,
                     "h": 35786400,
                     "rf": 298.257223563,
                     "proj": "geos",
                     "units": "m",
                     "sweep": "y"}

        # Asserts that the get_area_definition function was called with the correct arguments
        gad_.assert_called_once()
        args, kwargs = gad_.call_args
        assert args[0] == "mtg_fci_fdss_2km"
        assert args[1] == "MTG FCI Full Disk Scanning Service area definition with 2 km resolution"
        assert args[2] == ""
        assert args[3] == proj_dict
        assert args[4] == 10
        assert args[5] == 100

    def test_dataset(self):
        """Test the correct execution of the get_dataset function with a valid nc_key."""
        dataset = self.fh.get_dataset(make_dataid(name="test_one_layer", resolution=2000),
                                      {"name": "test_one_layer",
                                       "nc_key": "test_one_layer",
                                       "fill_value": -999,
                                       "file_type": "test_file_type"})

        np.testing.assert_allclose(dataset.values, np.arange(1000).reshape((100, 10)))
        assert dataset.attrs["test_attr"] == "attr"
        assert dataset.attrs["fill_value"] == -999

    def test_dataset_with_layer(self):
        """Check the correct execution of the get_dataset function with a valid nc_key & layer."""
        dataset = self.fh.get_dataset(make_dataid(name="test_two_layers", resolution=2000),
                                      {"name": "test_two_layers",
                                       "nc_key": "test_two_layers", "layer": 1,
                                       "fill_value": -999,
                                       "file_type": "test_file_type"})
        np.testing.assert_allclose(dataset.values, 2 * np.ones((100, 10)))
        assert dataset.attrs["spacecraft_name"] == "TEST_PLATFORM"

    def test_dataset_with_invalid_filekey(self):
        """Test the correct execution of the get_dataset function with an invalid nc_key."""
        invalid_dataset = self.fh.get_dataset(make_dataid(name="test_invalid", resolution=2000),
                                              {"name": "test_invalid",
                                               "nc_key": "test_invalid",
                                               "fill_value": -999,
                                               "file_type": "test_file_type"})
        assert invalid_dataset is None

    def test_dataset_with_total_cot(self):
        """Test the correct execution of the get_dataset function for total COT (add contributions from two layers)."""
        dataset = self.fh.get_dataset(make_dataid(name="retrieved_cloud_optical_thickness", resolution=2000),
                                      {"name": "retrieved_cloud_optical_thickness",
                                       "nc_key": "test_two_layers",
                                       "fill_value": -999,
                                       "file_type": "test_file_type"})
        # Checks that the function returns None
        expected_sum = np.empty((100, 10))
        expected_sum[:] = np.log10(10**2 + 10**1)
        np.testing.assert_allclose(dataset.values, expected_sum)

    def test_dataset_with_scalar(self):
        """Test the execution of the get_dataset function for scalar values."""
        # Checks returned scalar value
        dataset = self.fh.get_dataset(make_dataid(name="test_scalar"),
                                      {"name": "product_quality",
                                       "nc_key": "product_quality",
                                       "file_type": "test_file_type"})
        assert dataset.values == 99.0

        # Checks that no AreaDefintion is implemented for scalar values
        with pytest.raises(NotImplementedError):
            self.fh.get_area_def(None)

    def test_emumerations(self):
        """Test the conversion of enumerated type information into flag_values and flag_meanings."""
        dataset = self.fh.get_dataset(make_dataid(name="test_enum", resolution=2000),
                                      {"name": "quality_flag",
                                       "nc_key": "quality_flag",
                                       "file_type": "test_file_type",
                                       "import_enum_information": True})
        attributes = dataset.attrs
        assert "flag_values" in attributes
        assert attributes["flag_values"] == [0,1]
        assert "flag_meanings" in attributes
        assert attributes["flag_meanings"] == ["False","True"]

    def test_units_from_file(self):
        """Test units extraction from NetCDF file."""
        dataset = self.fh.get_dataset(make_dataid(name="test_units_from_file", resolution=2000),
                                      {"name": "test_one_layer",
                                       "nc_key": "test_one_layer",
                                       "file_type": "test_file_type"})
        assert dataset.attrs["units"] == "test_units"

    def test_unit_from_file(self):
        """Test that a unit stored with attribute `unit` in the file is assigned to the `units` attribute."""
        dataset = self.fh.get_dataset(make_dataid(name="test_unit_from_file", resolution=2000),
                                      {"name": "test_two_layers",
                                       "nc_key": "test_two_layers", "layer": 1,
                                       "file_type": "test_file_type"})
        assert dataset.attrs["units"] == "test_unit"

    def test_units_from_yaml(self):
        """Test units extraction from yaml file."""
        dataset = self.fh.get_dataset(make_dataid(name="test_units_from_yaml", resolution=2000),
                                      {"name": "test_one_layer",
                                       "units": "test_unit_from_yaml",
                                       "nc_key": "test_one_layer",
                                       "file_type": "test_file_type"})
        assert dataset.attrs["units"] == "test_unit_from_yaml"

    def test_units_none_conversion(self):
        """Test that a units stored as 'none' is converted to None."""
        dataset = self.fh.get_dataset(make_dataid(name="test_units_none_conversion", resolution=2000),
                                      {"name": "test_one_layer",
                                       "units": "none",
                                       "nc_key": "test_one_layer",
                                       "file_type": "test_file_type"})
        assert dataset.attrs["units"] is None


class TestFciL2NCSegmentFileHandler(unittest.TestCase):
    """Test the FciL2NCSegmentFileHandler reader."""

    def setUp(self):
        """Set up the test by creating a test file and opening it with the reader."""
        # Easiest way to test the reader is to create a test netCDF file on the fly
        self.seg_test_file = str(uuid.uuid4()) + ".nc"
        with Dataset(self.seg_test_file, "w") as nc:
            # Create dimensions
            nc.createDimension("number_of_FoR_cols", 348)
            nc.createDimension("number_of_FoR_rows", 348)
            nc.createDimension("number_of_channels", 8)
            nc.createDimension("number_of_categories", 6)

            # add global attributes
            nc.data_source = "TEST_FCI_DATA_SOURCE"
            nc.platform = "TEST_FCI_PLATFORM"

            # Add datasets
            x = nc.createVariable("x", np.float32, dimensions=("number_of_FoR_cols",))
            x.standard_name = "projection_x_coordinate"
            x[:] = np.arange(348)

            y = nc.createVariable("y", np.float32, dimensions=("number_of_FoR_rows",))
            y.standard_name = "projection_y_coordinate"
            y[:] = np.arange(348)

            s = nc.createVariable("product_quality", np.int8)
            s[:] = 99.

            chans = nc.createVariable("channels", np.float32, dimensions=("number_of_channels",))
            chans.standard_name = "fci_channels"
            chans[:] = np.arange(8)

            cats = nc.createVariable("categories", np.float32, dimensions=("number_of_categories",))
            cats.standard_name = "product_categories"
            cats[:] = np.arange(6)

            test_dataset = nc.createVariable("test_values", np.float32,
                                             dimensions=("number_of_FoR_rows", "number_of_FoR_cols",
                                                         "number_of_channels", "number_of_categories"))

            test_dataset[:] = self._get_unique_array(range(8), range(6))
            test_dataset.test_attr = "attr"
            test_dataset.units = "test_units"

    def tearDown(self):
        """Remove the previously created test file."""
        # First delete the fh, forcing the file to be closed if still open
        del self.fh
        # Then can safely remove it from the system
        with suppress(OSError):
            os.remove(self.seg_test_file)

    def test_all_basic(self):
        """Test all basic functionalities."""
        self.fh = FciL2NCSegmentFileHandler(filename=self.seg_test_file, filename_info={}, filetype_info={})

        assert self.fh.spacecraft_name == "TEST_FCI_PLATFORM"
        assert self.fh.sensor_name == "test_fci_data_source"
        assert self.fh.ssp_lon == 0.0

        global_attributes = self.fh._get_global_attributes()

        expected_global_attributes = {
            "filename": self.seg_test_file,
            "spacecraft_name": "TEST_FCI_PLATFORM",
            "ssp_lon": 0.0,
            "sensor": "test_fci_data_source",
            "platform_name": "TEST_FCI_PLATFORM"
        }
        assert global_attributes == expected_global_attributes

    def test_dataset(self):
        """Test the correct execution of the get_dataset function with valid nc_key."""
        self.fh = FciL2NCSegmentFileHandler(filename=self.seg_test_file, filename_info={}, filetype_info={})

        # Checks the correct execution of the get_dataset function with a valid nc_key
        dataset = self.fh.get_dataset(make_dataid(name="test_values", resolution=32000),
                                      {"name": "test_values",
                                       "nc_key": "test_values",
                                       "fill_value": -999, })
        expected_dataset = self._get_unique_array(range(8), range(6))
        np.testing.assert_allclose(dataset.values, expected_dataset)
        assert dataset.attrs["test_attr"] == "attr"
        assert dataset.attrs["units"] == "test_units"
        assert dataset.attrs["fill_value"] == -999

        # Checks that no AreaDefintion is implemented
        with pytest.raises(NotImplementedError):
            self.fh.get_area_def(None)

    def test_dataset_with_invalid_filekey(self):
        """Test the correct execution of the get_dataset function with an invalid nc_key."""
        self.fh = FciL2NCSegmentFileHandler(filename=self.seg_test_file, filename_info={}, filetype_info={})

        # Checks the correct execution of the get_dataset function with an invalid nc_key
        invalid_dataset = self.fh.get_dataset(make_dataid(name="test_invalid", resolution=32000),
                                              {"name": "test_invalid",
                                               "nc_key": "test_invalid",
                                               "fill_value": -999, })
        # Checks that the function returns None
        assert invalid_dataset is None

    def test_dataset_with_adef(self):
        """Test the correct execution of the get_dataset function with `with_area_definition=True`."""
        self.fh = FciL2NCSegmentFileHandler(filename=self.seg_test_file, filename_info={}, filetype_info={},
                                            with_area_definition=True)

        # Checks the correct execution of the get_dataset function with a valid nc_key
        dataset = self.fh.get_dataset(make_dataid(name="test_values", resolution=32000),
                                      {"name": "test_values",
                                       "nc_key": "test_values",
                                       "fill_value": -999,
                                       "coordinates": ("test_lon", "test_lat"), })
        expected_dataset = self._get_unique_array(range(8), range(6))
        np.testing.assert_allclose(dataset.values, expected_dataset)
        assert dataset.attrs["test_attr"] == "attr"
        assert dataset.attrs["units"] == "test_units"
        assert dataset.attrs["fill_value"] == -999

        # Checks returned AreaDefinition against reference
        adef = self.fh.get_area_def(None)
        assert adef == SEG_AREA_DEF

    def test_dataset_with_adef_and_wrongs_dims(self):
        """Test the correct execution of the get_dataset function with dims that don't match expected AreaDefinition."""
        self.fh = FciL2NCSegmentFileHandler(filename=self.seg_test_file, filename_info={}, filetype_info={},
                                            with_area_definition=True)
        with pytest.raises(NotImplementedError):
            self.fh.get_dataset(make_dataid(name="test_wrong_dims", resolution=6000),
                                {"name": "test_wrong_dims", "nc_key": "test_values", "fill_value": -999}
                                )

    def test_dataset_with_scalar(self):
        """Test the execution of the get_dataset function for scalar values."""
        self.fh = FciL2NCSegmentFileHandler(filename=self.seg_test_file, filename_info={}, filetype_info={})
        # Checks returned scalar value
        dataset = self.fh.get_dataset(make_dataid(name="test_scalar"),
                                      {"name": "product_quality",
                                       "nc_key": "product_quality",
                                       "file_type": "test_file_type"})
        assert dataset.values == 99.0

        # Checks that no AreaDefintion is implemented for scalar values
        with pytest.raises(NotImplementedError):
            self.fh.get_area_def(None)

    def test_dataset_slicing_catid(self):
        """Test the correct execution of the _slice_dataset function with 'category_id' set."""
        self.fh = FciL2NCSegmentFileHandler(filename=self.seg_test_file, filename_info={}, filetype_info={})

        dataset = self.fh.get_dataset(make_dataid(name="test_values", resolution=32000),
                                      {"name": "test_values",
                                       "nc_key": "test_values",
                                       "fill_value": -999,
                                       "category_id": 5})
        expected_dataset = self._get_unique_array(range(8), 5)
        np.testing.assert_allclose(dataset.values, expected_dataset)

    def test_dataset_slicing_chid_catid(self):
        """Test the correct execution of the _slice_dataset function with 'channel_id' and 'category_id' set."""
        self.fh = FciL2NCSegmentFileHandler(filename=self.seg_test_file, filename_info={}, filetype_info={})

        dataset = self.fh.get_dataset(make_dataid(name="test_values", resolution=32000),
                                      {"name": "test_values",
                                       "nc_key": "test_values",
                                       "fill_value": -999,
                                       "channel_id": 0, "category_id": 1})
        expected_dataset = self._get_unique_array(0, 1)
        np.testing.assert_allclose(dataset.values, expected_dataset)

    def test_dataset_slicing_visid_catid(self):
        """Test the correct execution of the _slice_dataset function with 'vis_channel_id' and 'category_id' set."""
        self.fh = FciL2NCSegmentFileHandler(filename=self.seg_test_file, filename_info={}, filetype_info={})

        self.fh.nc = self.fh.nc.rename_dims({"number_of_channels": "number_of_vis_channels"})
        dataset = self.fh.get_dataset(make_dataid(name="test_values", resolution=32000),
                                      {"name": "test_values",
                                       "nc_key": "test_values",
                                       "fill_value": -999,
                                       "vis_channel_id": 3, "category_id": 3})
        expected_dataset = self._get_unique_array(3, 3)
        np.testing.assert_allclose(dataset.values, expected_dataset)

    def test_dataset_slicing_irid(self):
        """Test the correct execution of the _slice_dataset function with 'ir_channel_id' set."""
        self.fh = FciL2NCSegmentFileHandler(filename=self.seg_test_file, filename_info={}, filetype_info={})

        self.fh.nc = self.fh.nc.rename_dims({"number_of_channels": "number_of_ir_channels"})
        dataset = self.fh.get_dataset(make_dataid(name="test_values", resolution=32000),
                                      {"name": "test_values",
                                       "nc_key": "test_values",
                                       "fill_value": -999,
                                       "ir_channel_id": 4})
        expected_dataset = self._get_unique_array(4, range(6))
        np.testing.assert_allclose(dataset.values, expected_dataset)

    @staticmethod
    def _get_unique_array(iarr, jarr):
        if not hasattr(iarr, "__iter__"):
            iarr = [iarr]

        if not hasattr(jarr, "__iter__"):
            jarr = [jarr]

        array = np.zeros((348, 348, 8, 6))
        for i in iarr:
            for j in jarr:
                array[:, :, i, j] = (i * 10) + j

        array = array[:, :, list(iarr), :]
        array = array[:, :, :, list(jarr)]

        return np.squeeze(array)


class TestFciL2NCReadingByteData(unittest.TestCase):
    """Test the FciL2NCFileHandler when reading and extracting byte data."""

    def setUp(self):
        """Set up the test by creating a test file and opening it with the reader."""
        # Easiest way to test the reader is to create a test netCDF file on the fly
        self.test_byte_file = str(uuid.uuid4()) + ".nc"
        with Dataset(self.test_byte_file, "w") as nc_byte:
            # Create dimensions
            nc_byte.createDimension("number_of_columns", 2)
            nc_byte.createDimension("number_of_rows", 2)

            # add global attributes
            nc_byte.data_source = "TEST_DATA_SOURCE"
            nc_byte.platform = "TEST_PLATFORM"

            # Add datasets
            x = nc_byte.createVariable("x", np.float32, dimensions=("number_of_columns",))
            x.standard_name = "projection_x_coordinate"
            x[:] = np.arange(2.0).reshape((2,))

            y = nc_byte.createVariable("y", np.float32, dimensions=("number_of_rows",))
            x.standard_name = "projection_y_coordinate"
            y[:] = np.arange(2.0).reshape((2,))

            mtg_geos_projection = nc_byte.createVariable("mtg_geos_projection", int, dimensions=())
            mtg_geos_projection.longitude_of_projection_origin = 0.0
            mtg_geos_projection.semi_major_axis = 6378137.
            mtg_geos_projection.inverse_flattening = 298.257223563
            mtg_geos_projection.perspective_point_height = 35786400.

            test_dataset = nc_byte.createVariable("cloud_mask_test_flag", np.float32,
                                                  dimensions=("number_of_rows", "number_of_columns",))

            # This number was chosen as we know the expected byte values
            test_dataset[:] = 4544767

        self.byte_reader = FciL2NCFileHandler(
            filename=self.test_byte_file,
            filename_info={},
            filetype_info={}
        )

    def tearDown(self):
        """Remove the previously created test file."""
        # First delete the file handler, forcing the file to be closed if still open
        del self.byte_reader
        # Then can safely remove it from the system
        with suppress(OSError):
            os.remove(self.test_byte_file)

    def test_byte_extraction(self):
        """Test the execution of the get_dataset function."""
        # Value of 1 is expected to be returned for this test
        dataset = self.byte_reader.get_dataset(make_dataid(name="cloud_mask_test_flag", resolution=2000),
                                               {"name": "cloud_mask_test_flag",
                                                "nc_key": "cloud_mask_test_flag",
                                                "fill_value": -999,
                                                "file_type": "nc_fci_test_clm",
                                                "extract_byte": 1,
                                                })

        np.testing.assert_allclose(dataset.values, 1)

        # Value of 0 is expected fto be returned or this test
        dataset = self.byte_reader.get_dataset(make_dataid(name="cloud_mask_test_flag", resolution=2000),
                                               {"name": "cloud_mask_test_flag",
                                                "nc_key": "cloud_mask_test_flag",
                                                "fill_value": -999, "mask_value": 0.,
                                                "file_type": "nc_fci_test_clm",
                                                "extract_byte": 23,
                                                })

        np.testing.assert_allclose(dataset.values, 0)


@pytest.fixture(scope="module")
def amv_file(tmp_path_factory):
    """Create an AMV file."""
    test_file = tmp_path_factory.mktemp("data") / "fci_l2_amv.nc"

    with Dataset(test_file, "w") as nc:
        # Create dimensions
        nc.createDimension("number_of_winds", 50000)

        # add global attributes
        nc.data_source = "TEST_DATA_SOURCE"
        nc.platform = "TEST_PLATFORM"

        # Add datasets
        latitude = nc.createVariable("latitude", np.float32, dimensions=("number_of_winds",))
        latitude[:] = np.arange(50000)

        longitude = nc.createVariable("y", np.float32, dimensions=("number_of_winds",))
        longitude[:] = np.arange(50000)

        qi = nc.createVariable("product_quality", np.int8)
        qi[:] = 99.

        test_dataset = nc.createVariable("test_dataset", np.float32,
                                                dimensions="number_of_winds")
        test_dataset[:] = np.ones(50000)
        test_dataset.test_attr = "attr"
        test_dataset.units = "test_units"

        mtg_geos_projection = nc.createVariable("mtg_geos_projection", int, dimensions=())
        mtg_geos_projection.longitude_of_projection_origin = 0.0
        mtg_geos_projection.semi_major_axis = 6378137.
        mtg_geos_projection.inverse_flattening = 298.257223563
        mtg_geos_projection.perspective_point_height = 35786400.
    return test_file


@pytest.fixture(scope="module")
def amv_filehandler(amv_file):
    """Create an AMV filehandler."""
    fh = FciL2NCAMVFileHandler(filename=amv_file,
                                 filename_info={"channel":"test_channel"},
                                 filetype_info={}
                                )
    mock_wind_time_ = xr.DataArray(
        np.array(8.14275e+08),  # seconds since 2000-01-01
        name="wind_time"
    )
    fh.nc = fh.nc.assign(wind_time=mock_wind_time_)
    return fh

class TestFciL2NCAMVFileHandler:
    """Test the FciL2NCAMVFileHandler reader."""

    def test_all_basic(self, amv_filehandler, amv_file):
        """Test all basic functionalities."""
        assert amv_filehandler.spacecraft_name == "TEST_PLATFORM"
        assert amv_filehandler.sensor_name == "test_data_source"
        assert amv_filehandler.ssp_lon == 0.0

        global_attributes = amv_filehandler._get_global_attributes(product_type="amv")
        expected_global_attributes = {
            "filename": amv_file,
            "spacecraft_name": "TEST_PLATFORM",
            "sensor": "test_data_source",
            "platform_name": "TEST_PLATFORM",
            "channel": "test_channel",
            "ssp_lon": 0.0,
            "time_parameters": {"wind_time":dt.datetime(2025, 10, 20, 11, 30)}
        }
        assert global_attributes == expected_global_attributes

        # Drop wind_time dataset and check that wind_time attribute becomes None
        amv_filehandler.nc = amv_filehandler.nc.drop_vars("wind_time")
        expected_global_attributes["time_parameters"] = {"wind_time":None}
        global_attributes = amv_filehandler._get_global_attributes(product_type="amv")
        assert global_attributes == expected_global_attributes

    def test_dataset(self, amv_filehandler):
        """Test the correct execution of the get_dataset function with a valid nc_key."""
        dataset = amv_filehandler.get_dataset(make_dataid(name="test_dataset", resolution=2000),
                                      {"name": "test_dataset",
                                       "nc_key": "test_dataset",
                                       "fill_value": -999,
                                       "file_type": "test_file_type"})
        np.testing.assert_allclose(dataset.values, np.ones(50000))
        assert dataset.attrs["test_attr"] == "attr"
        assert dataset.attrs["units"] == "test_units"
        assert dataset.attrs["fill_value"] == -999

    def test_dataset_with_invalid_filekey(self, amv_filehandler):
         """Test the correct execution of the get_dataset function with an invalid nc_key."""
         invalid_dataset = amv_filehandler.get_dataset(make_dataid(name="test_invalid", resolution=2000),
                                               {"name": "test_invalid",
                                                "nc_key": "test_invalid",
                                                "fill_value": -999,
                                                "file_type": "test_file_type"})
         assert invalid_dataset is None



# Copyright (c) 2015-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Composites."""

from __future__ import annotations

import logging

import dask.array as da
import numpy as np
import xarray as xr

from satpy.dataset import combine_metadata

from .core import CompositeBase, GenericCompositor, SingleBandCompositor

LOG = logging.getLogger(__name__)

MASKING_COMPOSITOR_METHODS = ["less", "less_equal", "equal", "greater_equal",
                              "greater", "not_equal", "isnan", "isfinite",
                              "isneginf", "isposinf"]


class MaskingCompositor(GenericCompositor):
    """A compositor that masks e.g. IR 10.8 channel data using cloud products from NWC SAF."""

    _supported_modes = {"LA", "RGBA"}

    def __init__(self, name, transparency=None, conditions=None, mode="LA",
                 **kwargs):
        """Collect custom configuration values.

        Kwargs:
            transparency (dict): transparency for each cloud type as
                                 key-value pairs in a dictionary.
                                 Will be converted to `conditions`.
                                 DEPRECATED.
            conditions (list): list of three items determining the masking
                               settings.
            mode (str, optional): Image mode to return.  For single-band input,
                                  this shall be "LA" (default) or "RGBA".  For
                                  multi-band input, this argument is ignored
                                  as the result is always RGBA.

        Each condition in *conditions* consists of three items:

        - `method`: Numpy method name.  The following are supported
            operations: `less`, `less_equal`, `equal`, `greater_equal`,
            `greater`, `not_equal`, `isnan`, `isfinite`, `isinf`,
            `isneginf`, or `isposinf`.
        - `value`: threshold value of the *mask* applied with the
            operator.  Can be a string, in which case the corresponding
            value will be determined from `flag_meanings` and
            `flag_values` attributes of the mask.
            NOTE: the `value` should not be given to 'is*` methods.
        - `transparency`: transparency from interval [0 ... 100] used
            for the method/threshold. Value of 100 is fully transparent.

        Example::

          >>> conditions = [{'method': 'greater_equal', 'value': 0,
                             'transparency': 100},
                            {'method': 'greater_equal', 'value': 1,
                             'transparency': 80},
                            {'method': 'greater_equal', 'value': 2,
                             'transparency': 0},
                            {'method': 'isnan',
                             'transparency': 100}]
          >>> compositor = MaskingCompositor("masking compositor",
                                             transparency=transparency)
          >>> result = compositor([data, mask])


        This will set transparency of `data` based on the values in
        the `mask` dataset.  Locations where `mask` has values of `0`
        will be fully transparent, locations with `1` will be
        semi-transparent and locations with `2` will be fully visible
        in the resulting image.  In the end all `NaN` areas in the mask are
        set to full transparency.  All the unlisted locations will be
        visible.

        The transparency is implemented by adding an alpha layer to
        the composite.  The locations with transparency of `100` will
        be set to NaN in the data.  If the input `data` contains an
        alpha channel, it will be discarded.

        """
        if transparency:
            LOG.warning("Using 'transparency' is deprecated in "
                        "MaskingCompositor, use 'conditions' instead.")
            self.conditions = []
            for key, transp in transparency.items():
                self.conditions.append({"method": "equal",
                                        "value": key,
                                        "transparency": transp})
            LOG.info("Converted 'transparency' to 'conditions': %s",
                     str(self.conditions))
        else:
            self.conditions = conditions
        if self.conditions is None:
            raise ValueError("Masking conditions not defined.")
        if mode not in self._supported_modes:
            raise ValueError(f"Invalid mode {mode!s}.  Supported modes: " +
                             ", ".join(self._supported_modes))
        self.mode = mode

        super(MaskingCompositor, self).__init__(name, **kwargs)

    def __call__(self, projectables, *args, **kwargs):
        """Call the compositor."""
        if len(projectables) != 2:
            raise ValueError("Expected 2 datasets, got %d" % (len(projectables),))
        projectables = self.match_data_arrays(projectables)
        data_in = projectables[0]
        mask_in = projectables[1]

        # remove "bands" dimension for single band masks (ex. "L")
        mask_in = mask_in.squeeze(drop=True)

        alpha_attrs = data_in.attrs.copy()
        data = self._select_data_bands(data_in)

        alpha = self._get_alpha_bands(data, mask_in, alpha_attrs)
        data.append(alpha)

        res = super(MaskingCompositor, self).__call__(data, **kwargs)
        return res

    def _get_mask(self, method, value, mask_data):
        """Get mask array from *mask_data* using *method* and threshold *value*.

        The *method* is the name of a numpy function.

        """
        if method not in MASKING_COMPOSITOR_METHODS:
            raise AttributeError("Unsupported Numpy method %s, use one of %s",
                                 method, str(MASKING_COMPOSITOR_METHODS))

        func = getattr(np, method)

        if value is None:
            return func(mask_data)
        return func(mask_data, value)

    def _set_data_nans(self, data, mask, attrs):
        """Set *data* to nans where *mask* is True.

        The attributes *attrs** will be written to each band in *data*.

        """
        for i, dat in enumerate(data):
            data[i] = xr.where(mask, np.nan, dat)
            data[i].attrs = attrs

        return data

    def _select_data_bands(self, data_in):
        """Select data to be composited from input data.

        From input data, select the bands that need to have masking applied.
        """
        if "bands" in data_in.dims:
            return [data_in.sel(bands=b) for b in data_in["bands"] if b != "A"]
        if self.mode == "RGBA":
            return [data_in, data_in, data_in]
        return [data_in]

    def _get_alpha_bands(self, data, mask_in, alpha_attrs):
        """Get alpha bands.

        From input data, masks, and attributes, get alpha band.
        """
        # Create alpha band
        mask_data = mask_in.data
        alpha = da.ones((data[0].sizes["y"],
                         data[0].sizes["x"]),
                        chunks=data[0].chunks)

        for condition in self.conditions:
            method = condition["method"]
            value = condition.get("value", None)
            if isinstance(value, str):
                value = _get_flag_value(mask_in, value)
            transparency = condition["transparency"]
            mask = self._get_mask(method, value, mask_data)

            if transparency == 100.0:
                data = self._set_data_nans(data, mask, alpha_attrs)
            alpha_val = 1. - transparency / 100.
            alpha = da.where(mask, alpha_val, alpha)

        return xr.DataArray(data=alpha, attrs=alpha_attrs,
                            dims=data[0].dims, coords=data[0].coords)


def _get_flag_value(mask, val):
    """Get a numerical value of the named flag.

    This function assumes the naming used in product generated with
    NWC SAF GEO/PPS softwares.

    """
    flag_meanings = mask.attrs["flag_meanings"]
    flag_values = mask.attrs["flag_values"]
    if isinstance(flag_meanings, str):
        flag_meanings = flag_meanings.split()

    index = flag_meanings.index(val)

    return flag_values[index]


class LongitudeMaskingCompositor(SingleBandCompositor):
    """Masks areas outside defined longitudes."""

    def __init__(self, name, lon_min=None, lon_max=None, **kwargs):  # noqa: D417
        """Collect custom configuration values.

        Args:
            lon_min (float): lower longitude limit
            lon_max (float): upper longitude limit
        """
        self.lon_min = lon_min
        self.lon_max = lon_max
        if self.lon_min is None and self.lon_max is None:
            raise ValueError("Masking conditions not defined. \
                At least lon_min or lon_max has to be specified.")
        if not self.lon_min:
            self.lon_min = -180.
        if not self.lon_max:
            self.lon_max = 180.
        super().__init__(name, **kwargs)

    def __call__(self, projectables, nonprojectables=None, **info):
        """Generate the composite."""
        projectable = projectables[0]
        lons, lats = projectable.attrs["area"].get_lonlats()

        if self.lon_max > self.lon_min:
            lon_min_max = np.logical_and(lons >= self.lon_min, lons <= self.lon_max)
        else:
            lon_min_max = np.logical_or(lons >= self.lon_min, lons <= self.lon_max)

        masked_projectable = projectable.where(lon_min_max)
        return super().__call__([masked_projectable], **info)


class SimpleFireMaskCompositor(CompositeBase):
    """Class for a simple fire detection compositor."""

    def __call__(self, projectables, nonprojectables=None, **attrs):
        """Compute a simple fire detection to create a boolean mask to be used in "flames" composites.

        Expects 4 channel inputs, calibrated to BT/reflectances, in this order [µm]: 10.x, 3.x, 2.x, 0.6.

        It applies 4 spectral tests, for which the thresholds must be provided in the yaml as "test_thresholds":
        - Test 0: 10.x > thr0 (clouds filter)
        - Test 1: 3.x-10.x > thr1 (hotspot)
        - Test 2: 0.6 > thr2 (clouds, sunglint filter)
        - Test 3: 3.x+2.x > thr3 (hotspot)

        .. warning::
            This fire detection algorithm is extremely simple, so it is prone to false alarms and missed detections.
            It is intended only for PR-like visualisation of large fires, not for any other use.
            The tests have been designed for MTG-FCI.

        """
        projectables = self.match_data_arrays(projectables)
        info = combine_metadata(*projectables)
        info["name"] = self.attrs["name"]
        info.update(self.attrs)

        # fire spectral tests

        # test 0: # window channel should be warm (no clouds)
        ir_105_temp = projectables[0] > self.attrs["test_thresholds"][0]
        # test 1: # 3.8-10.5µm should be high (hotspot)
        temp_diff = projectables[1] - projectables[0] > self.attrs["test_thresholds"][1]
        # test 2: vis_06 should be low (no clouds, no sunglint)
        vis_06_bright = projectables[3] < self.attrs["test_thresholds"][2]
        # test 3: 3.8+2.2µm should be high (hotspot)
        ir38_plus_nir22 = projectables[1] + projectables[2] >= self.attrs["test_thresholds"][3]

        res = ir_105_temp & temp_diff & vis_06_bright & ir38_plus_nir22  # combine all tests

        res.attrs = info
        return res


class CloudCompositor(GenericCompositor):
    """Detect clouds based on thresholding and use it as a mask for compositing."""

    def __init__(self, name, transition_min=258.15, transition_max=298.15,  # noqa: D417
                 transition_gamma=3.0, invert_alpha=False, **kwargs):
        """Collect custom configuration values.

        Args:
            transition_min (float): Values below or equal to this are
                                    clouds -> opaque white
            transition_max (float): Values above this are
                                    cloud free -> transparent
            transition_gamma (float): Gamma correction to apply at the end
            invert_alpha (bool): Invert the alpha channel to make low data values transparent
                                 and high data values opaque.

        """
        self.transition_min = transition_min
        self.transition_max = transition_max
        self.transition_gamma = transition_gamma
        self.invert_alpha = invert_alpha
        super(CloudCompositor, self).__init__(name, **kwargs)

    def __call__(self, projectables, **kwargs):
        """Generate the composite."""
        data = projectables[0]

        # Default to rough IR thresholds
        # Values below or equal to this are clouds -> opaque white
        tr_min = self.transition_min
        # Values above this are cloud free -> transparent
        tr_max = self.transition_max
        # Gamma correction
        gamma = self.transition_gamma

        slope = 1 / (tr_min - tr_max)
        offset = 1 - slope * tr_min

        alpha = data.where(data > tr_min, 1.)
        alpha = alpha.where(data <= tr_max, 0.)
        alpha = alpha.where((data <= tr_min) | (data > tr_max), slope * data + offset)

        if self.invert_alpha:
            alpha.data = 1.0 - alpha.data

        # gamma adjustment
        alpha **= gamma
        res = super(CloudCompositor, self).__call__((data, alpha), **kwargs)
        return res


class HighCloudCompositor(CloudCompositor):
    """Detect high clouds based on latitude-dependent thresholding and use it as a mask for compositing.

    This compositor aims at identifying high clouds and assigning them a transparency based on the brightness
    temperature (cloud opacity). In contrast to the `CloudCompositor`, the brightness temperature threshold at
    the lower end, used to identify high opaque clouds, is made a function of the latitude in order to have
    tropopause level clouds appear opaque at both high and low latitudes. This follows the Geocolor
    implementation of high clouds in Miller et al. (2020, :doi:`10.1175/JTECH-D-19-0134.1`), but
    with some adjustments to the thresholds based on recent developments and feedback from CIRA.

    The two brightness temperature thresholds in `transition_min` are used together with the corresponding
    latitude limits in `latitude_min` to compute a modified version of `transition_min` that is later used
    when calling `CloudCompositor`. The modified version of `transition_min` will be an array with the same
    shape as the input projectable dataset, where the actual values of threshold_min are a function of the
    dataset `latitude`:

      - transition_min = transition_min[0] where abs(latitude) < latitude_min(0)
      - transition_min = transition_min[1] where abs(latitude) > latitude_min(0)
      - transition_min = linear interpolation between transition_min[0] and transition_min[1] as a function
                         of where abs(latitude).
    """

    def __init__(self, name, transition_min_limits=(210., 230.), latitude_min_limits=(30., 60.),  # noqa: D417
                 transition_max=300, transition_gamma=1.0, **kwargs):
        """Collect custom configuration values.

        Args:
            transition_min_limits (tuple): Brightness temperature values used to identify opaque white
                                           clouds at different latitudes
            transition_max (float): Brightness temperatures above this value are not considered to
                                    be high clouds -> transparent
            latitude_min_limits (tuple): Latitude values defining the intervals for computing latitude-dependent
                                         `transition_min` values from `transition_min_limits`.
            transition_gamma (float): Gamma correction to apply to the alpha channel within the brightness
                                      temperature range (`transition_min` to `transition_max`).

        """
        if len(transition_min_limits) != 2:
            raise ValueError(f"Expected 2 `transition_min_limits` values, got {len(transition_min_limits)}")
        if len(latitude_min_limits) != 2:
            raise ValueError(f"Expected 2 `latitude_min_limits` values, got {len(latitude_min_limits)}")
        if type(transition_max) in [list, tuple]:
            raise ValueError(f"Expected `transition_max` to be of type float, is of type {type(transition_max)}")

        self.transition_min_limits = transition_min_limits
        self.latitude_min_limits = latitude_min_limits
        super().__init__(name, transition_min=None, transition_max=transition_max,
                         transition_gamma=transition_gamma, **kwargs)

    def __call__(self, projectables, **kwargs):
        """Generate the composite.

        `projectables` is expected to be a list or tuple with a single element:
          - index 0: Brightness temperature of a thermal infrared window channel (e.g. 10.5 microns).
        """
        if len(projectables) != 1:
            raise ValueError(f"Expected 1 dataset, got {len(projectables)}")

        data = projectables[0]
        _, lats = data.attrs["area"].get_lonlats(chunks=data.chunks, dtype=data.dtype)
        lats = np.abs(lats)

        slope = (self.transition_min_limits[1] - self.transition_min_limits[0]) / \
                (self.latitude_min_limits[1] - self.latitude_min_limits[0])
        offset = self.transition_min_limits[0] - slope * self.latitude_min_limits[0]

        # Compute pixel-level latitude dependent transition_min values and pass to parent CloudCompositor class
        transition_min = xr.DataArray(name="transition_min", coords=data.coords, dims=data.dims).astype(data.dtype)
        transition_min = transition_min.where(lats >= self.latitude_min_limits[0], self.transition_min_limits[0])
        transition_min = transition_min.where(lats <= self.latitude_min_limits[1], self.transition_min_limits[1])
        transition_min = transition_min.where((lats < self.latitude_min_limits[0]) |
                                              (lats > self.latitude_min_limits[1]), slope * lats + offset)
        self.transition_min = transition_min

        return super().__call__(projectables, **kwargs)


class LowCloudCompositor(CloudCompositor):
    """Detect low-level clouds based on thresholding and use it as a mask for compositing during night-time.

    This compositor computes the brightness temperature difference between a window channel (e.g. 10.5 micron)
    and the near-infrared channel e.g. (3.8 micron) and uses this brightness temperature difference, `BTD`, to
    create a partially transparent mask for compositing.

    Pixels with `BTD` values below a given threshold  will be transparent, whereas pixels with `BTD` values
    above another threshold will be opaque. The transparency of all other `BTD` values will be a linear
    function of the `BTD` value itself. Two sets of thresholds are used, one set for land surface types
    (`range_land`) and another one for water surface types (`range_water`), respectively. Hence,
    this compositor requires a land-water-mask as a prerequisite input. This follows the GeoColor
    implementation of night-time low-level clouds in Miller et al. (2020, :doi:`10.1175/JTECH-D-19-0134.1`), but
    with some adjustments to the thresholds based on recent developments and feedback from CIRA.

    Please note that the spectral test and thus the output of the compositor (using the expected input data) is
    only applicable during night-time.
    """

    def __init__(self, name, values_land=(1,), values_water=(0,),  # noqa: D417
                 range_land=(0.0, 4.0),
                 range_water=(0.0, 4.0),
                 transition_gamma=1.0,
                 invert_alpha=True, **kwargs):
        """Init info.

        Collect custom configuration values.

        Args:
            values_land (list): List of values used to identify land surface pixels in the land-water-mask.
            values_water (list): List of values used to identify water surface pixels in the land-water-mask.
            range_land (tuple): Threshold values used for masking low-level clouds from the brightness temperature
                                difference over land surface types.
            range_water (tuple): Threshold values used for masking low-level clouds from the brightness temperature
                                 difference over water.
            transition_gamma (float): Gamma correction to apply to the alpha channel within the brightness
                                      temperature difference range.
            invert_alpha (bool): Invert the alpha channel to make low data values transparent
                                 and high data values opaque.
        """
        if len(range_land) != 2:
            raise ValueError(f"Expected 2 `range_land` values, got {len(range_land)}")
        if len(range_water) != 2:
            raise ValueError(f"Expected 2 `range_water` values, got {len(range_water)}")

        self.values_land = values_land if type(values_land) in [list, tuple] else [values_land]
        self.values_water = values_water if type(values_water) in [list, tuple] else [values_water]
        self.range_land = range_land
        self.range_water = range_water
        super().__init__(name, transition_min=None, transition_max=None,
                         transition_gamma=transition_gamma, invert_alpha=invert_alpha, **kwargs)

    def __call__(self, projectables, **kwargs):
        """Generate the composite.

        `projectables` is expected to be a list or tuple with the following three elements:
          - index 0: Brightness temperature difference between a window channel (e.g. 10.5 micron) and a
                     near-infrared channel e.g. (3.8 micron).
          - index 1. Brightness temperature of the window channel (used to filter out noise-induced false alarms).
          - index 2: Land-Sea-Mask.
        """
        if len(projectables) != 3:
            raise ValueError(f"Expected 3 datasets, got {len(projectables)}")

        projectables = self.match_data_arrays(projectables)
        btd, bt_win, lsm = projectables
        lsm = lsm.squeeze(drop=True)
        lsm = lsm.round()  # Make sure to have whole numbers in case of smearing from resampling

        # Call CloudCompositor for land surface pixels
        self.transition_min, self.transition_max = self.range_land
        res = super().__call__([btd.where(lsm.isin(self.values_land))], **kwargs)

        # Call CloudCompositor for /water surface pixels
        self.transition_min, self.transition_max = self.range_water
        res_water = super().__call__([btd.where(lsm.isin(self.values_water))], **kwargs)

        # Compine resutls for land and water surface pixels
        res = res.where(lsm.isin(self.values_land), res_water)

        # Make pixels with cold window channel brightness temperatures transparent to avoid spurious false
        # alarms caused by noise in the 3.9um channel that can occur for very cold cloud tops
        res.loc["A"] = res.sel(bands="A").where(bt_win >= 230, 0.0)

        return res



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2022-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""VIIRS NOAA enterprise EDR product reader.

This module defines the :class:`VIIRSJRRFileHandler` file handler, to
be used for reading VIIRS EDR products generated by the NOAA enterprise
suite, which are downloadable via NOAA CLASS or on NOAA's AWS buckets.

A wide variety of such products exist and, at present, only a subset are supported.

 - Cloud mask: JRR-CloudMask_v2r3_j01_s202112250807275_e202112250808520_c202112250837300.nc
 - Cloud products: JRR-CloudHeight_v2r3_j01_s202112250807275_e202112250808520_c202112250837300.nc
 - Aerosol detection: JRR-ADP_v2r3_j01_s202112250807275_e202112250808520_c202112250839550.nc
 - Aerosol optical depth: JRR-AOD_v2r3_j01_s202112250807275_e202112250808520_c202112250839550.nc
 - Surface reflectance: SurfRefl_v1r1_j01_s202112250807275_e202112250808520_c202112250845080.nc
 - Land Surface Temperature: LST_v2r0_npp_s202307241724558_e202307241726200_c202307241854058.nc

All products use the same base reader ``viirs_edr`` and can be read through satpy with::

    import satpy
    import glob

    filenames = glob.glob('JRR-ADP*.nc')
    scene = satpy.Scene(filenames, reader='viirs_edr')
    scene.load(['smoke_concentration'])

.. note::

    Multiple products contain datasets with the same name! For example, both the cloud mask
    and aerosol detection files contain a cloud mask, but these are not identical.
    For clarity, the aerosol file cloudmask is named `cloud_mask_adp` in this reader.

Vegetation Indexes
^^^^^^^^^^^^^^^^^^

The NDVI and EVI products can be loaded from CSPP-produced Surface Reflectance
files. By default, these products are filtered based on the Surface Reflectance
Quality Flags. This is used to remove/mask pixels in certain cloud or water
regions. This behavior can be disabled by providing the reader keyword argument
``filter_veg`` and setting it to ``False``. For example::

    scene = satpy.Scene(filenames, reader='viirs_edr', reader_kwargs={"filter_veg": False})

AOD Filtering
^^^^^^^^^^^^^

The AOD (Aerosol Optical Depth) product can be optionally filtered based on
Quality Control (QC) values in the file. By default no filtering is performed.
By providing the ``aod_qc_filter`` keyword argument and specifying the maximum
value of the ``QCAll`` variable to include (not mask). For example::

    scene = satpy.Scene(filenames, reader='viirs_edr', reader_kwargs={"aod_qc_filter": 1})

will only preserve AOD550 values where the quality is 0 ("high") or
1 ("medium"). At the time of writing the ``QCAll`` variable has 1 ("medium"),
2 ("low"), and 3 ("no retrieval").


"""
from __future__ import annotations

import logging
from typing import Any, Iterable

import dask.array as da
import numpy as np
import xarray as xr

from satpy import DataID
from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.utils import get_chunk_size_limit

LOG = logging.getLogger(__name__)
M_COLS = 3200


class VIIRSJRRFileHandler(BaseFileHandler):
    """NetCDF4 reader for VIIRS Active Fires."""

    def __init__(self, filename, filename_info, filetype_info, **kwargs):
        """Initialize the geo filehandler."""
        super(VIIRSJRRFileHandler, self).__init__(filename, filename_info,
                                                  filetype_info)
        # use entire scans as chunks
        row_chunks_m = max(get_chunk_size_limit() // 4 // M_COLS, 1)  # 32-bit floats
        row_chunks_i = row_chunks_m * 2
        drop_variables = filetype_info.get("drop_variables", None)
        self.nc = xr.open_dataset(self.filename,
                                  decode_cf=filetype_info.get("decode_cf", True),
                                  mask_and_scale=True,
                                  drop_variables=drop_variables,
                                  chunks={
                                      "Columns": -1,
                                      "Rows": row_chunks_m,
                                      "Along_Scan_375m": -1,
                                      "Along_Track_375m": row_chunks_i,
                                      "Along_Scan_750m": -1,
                                      "Along_Track_750m": row_chunks_m,
                                  })

        # For some reason, no 'standard_name' is defined in some netCDF files, so
        # here we manually make the definitions.
        if "Latitude" in self.nc:
            self.nc["Latitude"].attrs.update({"standard_name": "latitude"})
        if "Longitude" in self.nc:
            self.nc["Longitude"].attrs.update({"standard_name": "longitude"})

        self.algorithm_version = filename_info["platform_shortname"]
        self.sensor_name = "viirs"

    def rows_per_scans(self, data_arr: xr.DataArray) -> int:
        """Get number of array rows per instrument scan based on data resolution."""
        return 16 if data_arr.shape[1] == M_COLS else 32

    def get_dataset(self, dataset_id: DataID, info: dict) -> xr.DataArray:
        """Get the dataset."""
        data_arr = self.nc[info["file_key"]]
        data_arr = self._mask_invalid(data_arr, info)
        data_arr = self._sanitize_metadata(data_arr, info)
        units = info.get("units", data_arr.attrs.get("units"))
        if units is None or units == "unitless":
            units = "1"
        if units == "%" and data_arr.attrs.get("units") in ("1", "unitless"):
            data_arr *= 100.0  # turn into percentages
        data_arr.attrs["units"] = units
        if data_arr.attrs.get("standard_name") in ("longitude", "latitude"):
            # recursive swath definitions are a problem for the base reader right now
            # delete the coordinates here so the base reader doesn't try to
            # make a SwathDefinition
            data_arr = data_arr.reset_coords(drop=True)

        return self._rename_dims(data_arr)

    def _mask_invalid(self, data_arr: xr.DataArray, ds_info: dict) -> xr.DataArray:
        # xarray auto mask and scale handled any fills from the file
        valid_range = ds_info.get("valid_range", data_arr.attrs.get("valid_range"))
        if "valid_min" in data_arr.attrs and valid_range is None:
            valid_range = (data_arr.attrs["valid_min"], data_arr.attrs["valid_max"])
        if valid_range is not None:
            # NOTE: may modify attrs in place
            fill_value = self._handle_fill_value(data_arr)
            return data_arr.where((valid_range[0] <= data_arr) & (data_arr <= valid_range[1]), fill_value)
        return data_arr

    def _handle_fill_value(self, data_arr: xr.DataArray) -> Any:
        if "_FillValue" in data_arr.attrs and not np.issubdtype(data_arr.dtype, np.floating):
            fill_value = data_arr.attrs["_FillValue"]
            if hasattr(fill_value, "shape"):
                fill_value = fill_value.item()
                data_arr.attrs["_FillValue"] = fill_value
        else:
            # fill value is being overwritten with NaN
            data_arr.attrs.pop("_FillValue", None)
            fill_value = xr.core.dtypes.NA
        return fill_value

    def _sanitize_metadata(self, data_arr: xr.DataArray, info: dict) -> xr.DataArray:
        if "valid_range" in data_arr.attrs:
            # don't use numpy arrays for simple metadata
            data_arr.attrs["valid_range"] = tuple(data_arr.attrs["valid_range"])
        if "standard_name" in info:
            data_arr.attrs["standard_name"] = info["standard_name"]
        self._decode_flag_meanings(data_arr)
        data_arr.attrs["platform_name"] = self.platform_name
        data_arr.attrs["sensor"] = self.sensor_name
        data_arr.attrs["rows_per_scan"] = self.rows_per_scans(data_arr)
        return data_arr

    @staticmethod
    def _decode_flag_meanings(data_arr: xr.DataArray):
        flag_meanings = data_arr.attrs.get("flag_meanings", None)
        if isinstance(flag_meanings, str) and "\n" not in flag_meanings:
            # only handle CF-standard flag meanings
            data_arr.attrs["flag_meanings"] = [flag for flag in data_arr.attrs["flag_meanings"].split(" ")]

    @staticmethod
    def _rename_dims(data_arr: xr.DataArray) -> xr.DataArray:
        if "Columns" in data_arr.dims:
            data_arr = data_arr.rename({"Columns": "x", "Rows": "y"})
        if "Along_Track_375m" in data_arr.dims:
            data_arr = data_arr.rename({"Along_Scan_375m": "x", "Along_Track_375m": "y"})
        if "Along_Track_750m" in data_arr.dims:
            data_arr = data_arr.rename({"Along_Scan_750m": "x", "Along_Track_750m": "y"})
        return data_arr

    @property
    def start_time(self):
        """Get first date/time when observations were recorded."""
        return self.filename_info["start_time"]

    @property
    def end_time(self):
        """Get last date/time when observations were recorded."""
        return self.filename_info["end_time"]

    @property
    def platform_name(self):
        """Get platform name."""
        platform_path = self.filename_info["platform_shortname"]
        platform_dict = {"NPP": "Suomi-NPP",
                         "JPSS-1": "NOAA-20",
                         "SNPP": "Suomi-NPP",
                         "J01": "NOAA-20",
                         "N20": "NOAA-20",
                         "JPSS-2": "NOAA-21",
                         "J02": "NOAA-21",
                         "N21": "NOAA-21"}
        return platform_dict[platform_path.upper()]

    def available_datasets(self, configured_datasets=None):
        """Get information of available datasets in this file.

        Args:
            configured_datasets (list): Series of (bool or None, dict) in the
                same way as is returned by this method (see below). The bool
                is whether the dataset is available from at least one
                of the current file handlers. It can also be ``None`` if
                no file handler before us knows how to handle it.
                The dictionary is existing dataset metadata. The dictionaries
                are typically provided from a YAML configuration file and may
                be modified, updated, or used as a "template" for additional
                available datasets. This argument could be the result of a
                previous file handler's implementation of this method.

        Returns:
            Iterator of (bool or None, dict) pairs where dict is the
            dataset's metadata. If the dataset is available in the current
            file type then the boolean value should be ``True``, ``False``
            if we **know** about the dataset but it is unavailable, or
            ``None`` if this file object is not responsible for it.

        """
        # keep track of what variables the YAML has configured, so we don't
        # duplicate entries for them in the dynamic portion
        handled_var_names = set()
        for is_avail, ds_info in (configured_datasets or []):
            file_key = ds_info.get("file_key", ds_info["name"])
            # we must add all variables here even if another file handler has
            # claimed the variable. It could be another instance of this file
            # type and we don't want to add that variable dynamically if the
            # other file handler defined it by the YAML definition.
            handled_var_names.add(file_key)
            if is_avail is not None:
                # some other file handler said it has this dataset
                # we don't know any more information than the previous
                # file handler so let's yield early
                yield is_avail, ds_info
                continue
            if self.file_type_matches(ds_info["file_type"]) is None:
                # this is not the file type for this dataset
                yield None, ds_info
                continue
            yield file_key in self.nc, ds_info

        yield from self._dynamic_variables_from_file(handled_var_names)

    def _dynamic_variables_from_file(self, handled_var_names: set) -> Iterable[tuple[bool, dict]]:
        coords: dict[str, dict] = {}
        for is_avail, ds_info in self._generate_dynamic_metadata(self.nc.variables.keys(), coords):
            var_name = ds_info["file_key"]
            if var_name in handled_var_names and not ("longitude_" in var_name or "latitude_" in var_name):
                continue
            handled_var_names.add(var_name)
            yield is_avail, ds_info

        for coord_info in coords.values():
            yield True, coord_info

    def _generate_dynamic_metadata(self, variable_names: Iterable[str], coords: dict) -> Iterable[tuple[bool, dict]]:
        for var_name in variable_names:
            data_arr = self.nc[var_name]
            if data_arr.ndim != 2:
                # only 2D arrays supported at this time
                continue
            res = 750 if data_arr.shape[1] == M_COLS else 375
            ds_info = {
                "file_key": var_name,
                "file_type": self.filetype_info["file_type"],
                "name": var_name,
                "resolution": res,
                "coordinates": self._coord_names_for_resolution(res),
            }

            is_lon = "longitude" in var_name.lower()
            is_lat = "latitude" in var_name.lower()
            if not (is_lon or is_lat):
                yield True, ds_info
                continue

            ds_info["standard_name"] = "longitude" if is_lon else "latitude"
            ds_info["units"] = "degrees_east" if is_lon else "degrees_north"
            # recursive coordinate/SwathDefinitions are not currently handled well in the base reader
            del ds_info["coordinates"]
            yield True, ds_info

            # "standard" geolocation coordinate (assume shorter variable name is "better")
            new_name = self._coord_names_for_resolution(res)[int(not is_lon)]
            if new_name not in coords or len(var_name) < len(coords[new_name]["file_key"]):
                ds_info = ds_info.copy()
                ds_info["name"] = new_name
                coords[ds_info["name"]] = ds_info

    def _coord_names_for_resolution(self, res: int):
        ftype = self.filetype_info["file_type"]
        m_lon_name = f"longitude_{ftype}"
        m_lat_name = f"latitude_{ftype}"
        m_coords = (m_lon_name, m_lat_name)
        i_lon_name = f"longitude_i_{ftype}"
        i_lat_name = f"latitude_i_{ftype}"
        i_coords = (i_lon_name, i_lat_name)
        if res == 750:
            return m_coords
        else:
            return i_coords



class VIIRSSurfaceReflectanceWithVIHandler(VIIRSJRRFileHandler):
    """File handler for surface reflectance files with optional vegetation indexes."""

    def __init__(self, *args, filter_veg: bool = True, **kwargs) -> None:
        """Initialize file handler and keep track of vegetation index filtering."""
        super().__init__(*args, **kwargs)
        self._filter_veg = filter_veg

    def _mask_invalid(self, data_arr: xr.DataArray, ds_info: dict) -> xr.DataArray:
        new_data_arr = super()._mask_invalid(data_arr, ds_info)
        if ds_info["file_key"] in ("NDVI", "EVI") and self._filter_veg:
            good_mask = self._get_veg_index_good_mask()
            new_data_arr = new_data_arr.where(good_mask)
        return new_data_arr

    def _get_veg_index_good_mask(self) -> da.Array:
        # each mask array should be TRUE when pixels are UNACCEPTABLE
        qf1 = self.nc["QF1 Surface Reflectance"]
        # has_sun_glint = (qf1 & 0b11000000) > 0
        is_cloudy = (qf1 & 0b00001100) > 0  # mask everything but "confident clear"
        cloud_quality = (qf1 & 0b00000011) < 0b10

        qf2 = self.nc["QF2 Surface Reflectance"]
        has_snow_or_ice = (qf2 & 0b00100000) > 0
        has_cloud_shadow = (qf2 & 0b00001000) > 0
        water_mask = (qf2 & 0b00000111)
        has_water = (water_mask <= 0b010) | (water_mask == 0b101)  # shallow water, deep ocean, arctic

        qf7 = self.nc["QF7 Surface Reflectance"]
        has_aerosols = (qf7 & 0b00001100) > 0b1000  # high aerosol quantity
        adjacent_to_cloud = (qf7 & 0b00000010) > 0

        bad_mask = (
                # has_sun_glint |
                is_cloudy |
                cloud_quality |
                has_snow_or_ice |
                has_cloud_shadow |
                has_water |
                has_aerosols |
                adjacent_to_cloud
        )
        # upscale from M-band resolution to I-band resolution
        bad_mask_iband_dask = bad_mask.data.repeat(2, axis=1).repeat(2, axis=0)
        return ~bad_mask_iband_dask


class VIIRSLSTHandler(VIIRSJRRFileHandler):
    """File handler to handle LST file scale factor and offset weirdness."""

    _manual_scalings = {
        "VLST": ("LST_ScaleFact", "LST_Offset"),
        "emis_m15": ("LSE_ScaleFact", "LSE_Offset"),
        "emis_m16": ("LSE_ScaleFact", "LSE_Offset"),
        "emis_bbe": ("LSE_ScaleFact", "LSE_Offset"),
        "Satellite_Azimuth_Angle": ("AZI_ScaleFact", "AZI_Offset"),
    }

    def __init__(self, *args, **kwargs):
        """Initialize the file handler and unscale necessary variables."""
        super().__init__(*args, **kwargs)

        # Update variables with external scale factor and offset
        self._scale_data()

    def _scale_data(self):
        for var_name in list(self.nc.variables.keys()):
            if var_name not in self._manual_scalings:
                continue
            data_arr = self.nc[var_name]
            scale_factor = self.nc[self._manual_scalings[var_name][0]]
            add_offset = self.nc[self._manual_scalings[var_name][1]]
            data_arr.data = data_arr.data * scale_factor.data + add_offset.data
            self.nc[var_name] = data_arr


class VIIRSAODHandler(VIIRSJRRFileHandler):
    """File handler for AOD data files."""

    def __init__(self, *args, aod_qc_filter: int | None = None, **kwargs) -> None:
        """Initialize file handler and keep track of QC filtering."""
        super().__init__(*args, **kwargs)
        self._aod_qc_filter = aod_qc_filter

    def _mask_invalid(self, data_arr: xr.DataArray, ds_info: dict) -> xr.DataArray:
        new_data_arr = super()._mask_invalid(data_arr, ds_info)
        if self._aod_qc_filter is None or ds_info["name"] != "AOD550":
            return new_data_arr
        LOG.debug(f"Filtering AOD data to include quality <= {self._aod_qc_filter}")
        qc_all = self.nc["QCAll"]
        return new_data_arr.where(qc_all <= self._aod_qc_filter)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""The epic_l1b_h5 reader tests package."""

import os
import tempfile

import h5py
import numpy as np
import pytest

from satpy.readers.epic_l1b_h5 import CALIB_COEFS
from satpy.tests.utils import RANDOM_GEN

b317_data = RANDOM_GEN.uniform(low=0, high=5200, size=(100, 100))
b688_data = RANDOM_GEN.uniform(low=0, high=5200, size=(100, 100))
sza_data = RANDOM_GEN.uniform(low=0, high=100, size=(100, 100))
vaa_data = RANDOM_GEN.uniform(low=-180, high=180, size=(100, 100))
lon_data = RANDOM_GEN.uniform(low=-90, high=90, size=(100, 100))
lat_data = RANDOM_GEN.uniform(low=-180, high=180, size=(100, 100))
mas_data = RANDOM_GEN.choice([0, 1], size=(100, 100))


@pytest.fixture
def setup_hdf5_file(tmp_path):
    """Create temp hdf5 files."""
    fn = tmp_path / "epic_1b_20150613120251_03.h5"
    make_fake_hdf_epic(fn)
    return fn


def make_fake_hdf_epic(fname):
    """Make a fake HDF5 file for EPIC data testing."""
    fid = h5py.File(fname, "w")
    g1 = fid.create_group("Band317nm")
    g1.create_dataset("Image", shape=(100, 100), dtype=np.float32, data=b317_data)
    g2 = fid.create_group("Band688nm")
    g2.create_dataset("Image", shape=(100, 100), dtype=np.float32, data=b688_data)
    g3 = g2.create_group("Geolocation")
    g4 = g3.create_group("Earth")
    g4.create_dataset("SunAngleZenith", shape=(100, 100), dtype=np.float32, data=sza_data)
    g4.create_dataset("ViewAngleAzimuth", shape=(100, 100), dtype=np.float32, data=vaa_data)
    g4.create_dataset("Mask", shape=(100, 100), dtype=int, data=mas_data)
    g4.create_dataset("Latitude", shape=(100, 100), dtype=np.float32, data=lat_data)
    g4.create_dataset("Longitude", shape=(100, 100), dtype=np.float32, data=lon_data)

    fid.attrs.create("begin_time", "2015-06-13 12:00:37")
    fid.attrs.create("end_time", "2015-06-13 12:05:01")
    fid.close()


class TestEPICL1bReader:
    """Test the EPIC L1b HDF5 reader."""

    def _setup_h5(self, setup_hdf5_file):
        """Initialise reader for the tests."""
        from satpy.readers.core.loading import load_reader
        test_reader = load_reader(self.reader_configs)
        loadables = test_reader.select_files_from_pathnames([setup_hdf5_file])
        test_reader.create_filehandlers(loadables)

        return test_reader

    def setup_method(self):
        """Set up the tests."""
        from satpy._config import config_search_paths

        self.yaml_file = "epic_l1b_h5.yaml"

        self.filename_test = os.path.join(
            tempfile.gettempdir(),
            "epic_1b_20150613120251_03.h5",
        )
        self.reader_configs = config_search_paths(os.path.join("readers", self.yaml_file))

    def test_times(self, setup_hdf5_file):
        """Test start and end times load properly."""
        import datetime as dt

        test_reader = self._setup_h5(setup_hdf5_file)
        assert test_reader.start_time == dt.datetime(2015, 6, 13, 12, 0, 37)
        assert test_reader.end_time == dt.datetime(2015, 6, 13, 12, 5, 1)

    def test_counts_calibration(self, setup_hdf5_file):
        """Test that data is correctly calibrated."""
        from satpy.tests.utils import make_dsq

        test_reader = self._setup_h5(setup_hdf5_file)
        # Test counts calibration
        ds = test_reader.load([make_dsq(name="B317", calibration="counts")])
        np.testing.assert_allclose(ds["B317"].data, b317_data)

    def test_refl_calibration(self, setup_hdf5_file):
        """Test that data is correctly calibrated into reflectances."""
        from satpy.tests.utils import make_dsq

        test_reader = self._setup_h5(setup_hdf5_file)

        # Test conversion to reflectance
        ds = test_reader.load([make_dsq(name="B317", calibration="reflectance")])
        np.testing.assert_allclose(ds["B317"].data, b317_data * CALIB_COEFS["B317"] * 100., rtol=1e-5)

    def test_bad_calibration(self, setup_hdf5_file):
        """Test that error is raised if a bad calibration is used."""
        from satpy.tests.utils import make_dsq

        test_reader = self._setup_h5(setup_hdf5_file)

        # Test nonsense calibration
        with pytest.raises(KeyError):
            test_reader.load([make_dsq(name="B317", calibration="potatoes")])

    def test_load_ancillary(self, setup_hdf5_file):
        """Test that ancillary datasets load correctly."""
        from satpy.tests.utils import make_dsq

        test_reader = self._setup_h5(setup_hdf5_file)

        # Load sza
        ds = test_reader.load([make_dsq(name="solar_zenith_angle"),
                               make_dsq(name="satellite_azimuth_angle"),
                               make_dsq(name="latitude"),
                               make_dsq(name="longitude"),
                               make_dsq(name="earth_mask")])

        np.testing.assert_allclose(ds["solar_zenith_angle"].data, sza_data)
        np.testing.assert_allclose(ds["satellite_azimuth_angle"].data, vaa_data)
        np.testing.assert_allclose(ds["latitude"].data, lat_data)
        np.testing.assert_allclose(ds["longitude"].data, lon_data)
        np.testing.assert_allclose(ds["earth_mask"].data, mas_data)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Pytest configuration and setup functions."""
import pytest


def pytest_configure(config):
    """Set test configuration."""
    from satpy import aux_download
    aux_download.RUNNING_TESTS = True


def pytest_unconfigure(config):
    """Undo previous configurations."""
    from satpy import aux_download
    aux_download.RUNNING_TESTS = False


@pytest.fixture(scope="session")
def session_tmp_path(tmp_path_factory):
    """Generate a single temp path to use for the entire session."""
    return tmp_path_factory.mktemp("data")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2020 Satpy developers
#
# satpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# satpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with satpy.  If not, see <http://www.gnu.org/licenses/>.

"""The vii_base_nc reader tests package."""

import datetime
import os
import unittest
import uuid
from unittest import mock

import numpy as np
import pytest
import xarray as xr
from netCDF4 import Dataset

from satpy.readers.core.vii_nc import SCAN_ALT_TIE_POINTS, TIE_POINTS_FACTOR, ViiNCBaseFileHandler

TEST_FILE = "test_file_vii_base_nc.nc"


class TestViiNCBaseFileHandler(unittest.TestCase):
    """Test the ViiNCBaseFileHandler reader."""

    @mock.patch("satpy.readers.core.vii_nc.ViiNCBaseFileHandler._perform_geo_interpolation")
    def setUp(self, pgi_):
        """Set up the test."""
        # Easiest way to test the reader is to create a test netCDF file on the fly
        # uses a UUID to avoid permission conflicts during execution of tests in parallel
        self.test_file_name = TEST_FILE + str(uuid.uuid1()) + ".nc"

        with Dataset(self.test_file_name, "w") as nc:
            # Add global attributes
            nc.sensing_start_time_utc = "20170920173040.888"
            nc.sensing_end_time_utc = "20170920174117.555"
            nc.spacecraft = "test_spacecraft"
            nc.instrument = "test_instrument"

            # Create data group
            g1 = nc.createGroup("data")

            # Add dimensions to data group
            g1.createDimension("num_pixels", 10)
            g1.createDimension("num_lines", 100)

            # Create data/measurement_data group
            g1_1 = g1.createGroup("measurement_data")

            # Add dimensions to data/measurement_data group
            g1_1.createDimension("num_tie_points_act", 10)
            g1_1.createDimension("num_tie_points_alt", 100)

            # Add variables to data/measurement_data group
            tpw = g1_1.createVariable("tpw", np.float32, dimensions=("num_pixels", "num_lines"))
            tpw[:] = 1.
            tpw.test_attr = "attr"
            lon = g1_1.createVariable("longitude",
                                      np.float32,
                                      dimensions=("num_tie_points_act", "num_tie_points_alt"))
            lon[:] = 100.
            lat = g1_1.createVariable("latitude",
                                      np.float32,
                                      dimensions=("num_tie_points_act", "num_tie_points_alt"))
            lat[:] = 10.

            # Create quality group
            g2 = nc.createGroup("quality")

            # Add dimensions to quality group
            g2.createDimension("gap_items", 2)

            # Add variables to quality group
            var = g2.createVariable("duration_of_product", np.double, dimensions=())
            var[:] = 1.0
            var = g2.createVariable("duration_of_data_present", np.double, dimensions=())
            var[:] = 2.0
            var = g2.createVariable("duration_of_data_missing", np.double, dimensions=())
            var[:] = 3.0
            var = g2.createVariable("duration_of_data_degraded", np.double, dimensions=())
            var[:] = 4.0
            var = g2.createVariable("gap_start_time_utc", np.double, dimensions=("gap_items",))
            var[:] = [5.0, 6.0]
            var = g2.createVariable("gap_end_time_utc", np.double, dimensions=("gap_items",))
            var[:] = [7.0, 8.0]

        # Create longitude and latitude "interpolated" arrays
        interp_longitude = xr.DataArray(np.ones((10, 100)))
        interp_latitude = xr.DataArray(np.ones((10, 100)) * 2.)
        pgi_.return_value = (interp_longitude, interp_latitude)

        # Filename info valid for all readers
        filename_info = {
            "creation_time": datetime.datetime(year=2017, month=9, day=22,
                                               hour=22, minute=40, second=10),
            "sensing_start_time": datetime.datetime(year=2017, month=9, day=20,
                                                    hour=12, minute=30, second=30),
            "sensing_end_time": datetime.datetime(year=2017, month=9, day=20,
                                                  hour=18, minute=30, second=50)
        }
        self.filename_info = filename_info

        # Create a reader
        self.reader = ViiNCBaseFileHandler(
            filename=self.test_file_name,
            filename_info=filename_info,
            filetype_info={
                "cached_longitude": "data/measurement_data/longitude",
                "cached_latitude": "data/measurement_data/latitude"
            }
        )

        # Create a second reader where orthorectification and interpolation are inhibited
        # by means of the filetype_info flags
        self.reader_2 = ViiNCBaseFileHandler(
            filename=self.test_file_name,
            filename_info=filename_info,
            filetype_info={
                "cached_longitude": "data/measurement_data/longitude",
                "cached_latitude": "data/measurement_data/latitude",
                "interpolate": False,
                "orthorect": False
            },
            orthorect=True
        )

        # Create a third reader without defining cached latitude and longitude
        # by means of the filetype_info flags
        self.reader_3 = ViiNCBaseFileHandler(
            filename=self.test_file_name,
            filename_info=filename_info,
            filetype_info={},
            orthorect=True
        )

    def tearDown(self):
        """Remove the previously created test file."""
        # Catch Windows PermissionError for removing the created test file.
        try:
            os.remove(self.test_file_name)
        except OSError:
            pass

    def test_file_reading(self):
        """Test the file product reading."""
        # Checks that the basic functionalities are correctly executed
        expected_start_time = datetime.datetime(year=2017, month=9, day=20,
                                                hour=17, minute=30, second=40, microsecond=888000)
        assert self.reader.start_time == expected_start_time

        expected_end_time = datetime.datetime(year=2017, month=9, day=20,
                                              hour=17, minute=41, second=17, microsecond=555000)
        assert self.reader.end_time == expected_end_time

        assert self.reader.spacecraft_name == "test_spacecraft"
        assert self.reader.sensor == "test_instrument"
        assert self.reader.ssp_lon is None

        # Checks that the global attributes are correctly read
        expected_global_attributes = {
            "filename": self.test_file_name,
            "start_time": expected_start_time,
            "end_time": expected_end_time,
            "spacecraft_name": "test_spacecraft",
            "ssp_lon": None,
            "sensor": "test_instrument",
            "filename_start_time": datetime.datetime(year=2017, month=9, day=20,
                                                     hour=12, minute=30, second=30),
            "filename_end_time": datetime.datetime(year=2017, month=9, day=20,
                                                   hour=18, minute=30, second=50),
            "platform_name": "test_spacecraft",
            "quality_group": {
                "duration_of_product": 1.,
                "duration_of_data_present": 2.,
                "duration_of_data_missing": 3.,
                "duration_of_data_degraded": 4.,
                "gap_start_time_utc": (5., 6.),
                "gap_end_time_utc": (7., 8.)
            }
        }

        global_attributes = self.reader._get_global_attributes()
        # Since the global_attributes dictionary contains numpy arrays,
        # it is not possible to peform a simple equality test
        # Must iterate on all keys to confirm that the dictionaries are equal
        assert global_attributes.keys() == expected_global_attributes.keys()
        for key in expected_global_attributes:
            if key not in ["quality_group"]:
                # Quality check must be valid for both iterable and not iterable elements
                try:
                    equal = all(global_attributes[key] == expected_global_attributes[key])
                except (TypeError, ValueError):
                    equal = global_attributes[key] == expected_global_attributes[key]
                assert equal
            else:
                assert global_attributes[key].keys() == expected_global_attributes[key].keys()
                for inner_key in global_attributes[key]:
                    # Equality check must be valid for both iterable and not iterable elements
                    try:
                        equal = all(global_attributes[key][inner_key] == expected_global_attributes[key][inner_key])
                    except (TypeError, ValueError):
                        equal = global_attributes[key][inner_key] == expected_global_attributes[key][inner_key]
                    assert equal

    def test_start_end_time_additional_formats(self):
        """Test parsing additional datetime formats."""
        time_cases = [
            (
                "2017-09-20T17:30:40.888000",
                datetime.datetime(2017, 9, 20, 17, 30, 40, 888000),
                "2017-09-20T17:41:17.555000",
                datetime.datetime(2017, 9, 20, 17, 41, 17, 555000),
            ),
            (
                "2017-09-20 17:30:40",
                datetime.datetime(2017, 9, 20, 17, 30, 40),
                "2017-09-20 17:41:17",
                datetime.datetime(2017, 9, 20, 17, 41, 17),
            ),
        ]

        for start_str, expected_start, end_str, expected_end in time_cases:
            with Dataset(self.test_file_name, "r+") as nc:
                nc.sensing_start_time_utc = start_str
                nc.sensing_end_time_utc = end_str

            reader = ViiNCBaseFileHandler(
                filename=self.test_file_name,
                filename_info=self.filename_info,
                filetype_info={},
            )

            assert reader.start_time == expected_start
            assert reader.end_time == expected_end

    def test_bad_start_end_time(self):
        """Test parsing a bad datetime format."""
        with Dataset(self.test_file_name, "r+") as nc:
            nc.sensing_start_time_utc = "201709201730"
            nc.sensing_end_time_utc = "201709201740"

        reader = ViiNCBaseFileHandler(
            filename=self.test_file_name,
            filename_info=self.filename_info,
            filetype_info={},
        )

        with pytest.raises(ValueError, match="Unrecognized datetime format"):
            reader.start_time
        with pytest.raises(ValueError, match="Unrecognized datetime format"):
            reader.end_time

    @mock.patch("satpy.readers.core.vii_nc.tie_points_interpolation")
    @mock.patch("satpy.readers.core.vii_nc.tie_points_geo_interpolation")
    def test_functions(self, tpgi_, tpi_):
        """Test the functions."""
        with pytest.raises(NotImplementedError):
            self.reader._perform_orthorectification(mock.Mock(), mock.Mock())

        with pytest.raises(NotImplementedError):
            self.reader._perform_calibration(mock.Mock(), mock.Mock())

        # Checks that the _perform_interpolation function is correctly executed
        variable = xr.DataArray(
            dims=("y", "x"),
            name="test_name",
            attrs={
                "key_1": "value_1",
                "key_2": "value_2"
            },
            data=np.zeros((10, 100)),
        )
        tpi_.return_value = [xr.DataArray(
            dims=("num_tie_points_act", "num_tie_points_alt"),
            data=np.ones((10, 100))
        )]

        return_value = self.reader._perform_interpolation(variable)

        tpi_.assert_called_with([variable], SCAN_ALT_TIE_POINTS, TIE_POINTS_FACTOR)
        assert np.allclose(return_value, np.ones((10, 100)))
        assert return_value.attrs == {"key_1": "value_1", "key_2": "value_2"}
        assert return_value.name == "test_name"
        assert return_value.dims == ("num_pixels", "num_lines")

        # Checks that the _perform_geo_interpolation function is correctly executed
        variable_lon = xr.DataArray(
            dims=("y", "x"),
            name="test_lon",
            attrs={
                "key_1": "value_lon_1",
                "key_2": "value_lon_2"
            },
            data=np.zeros((10, 100))
        )
        variable_lat = xr.DataArray(
            dims=("y", "x"),
            name="test_lat",
            attrs={
                "key_1": "value_lat_1",
                "key_2": "value_lat_2"
            },
            data=np.ones((10, 100)) * 2.
        )

        tpgi_.return_value = (
            xr.DataArray(
                dims=("num_tie_points_act", "num_tie_points_alt"),
                data=np.ones((10, 100))
            ),
            xr.DataArray(
                dims=("num_tie_points_act", "num_tie_points_alt"),
                data=6 * np.ones((10, 100))
            )
        )

        return_lon, return_lat = self.reader._perform_geo_interpolation(variable_lon, variable_lat)

        tpgi_.assert_called_with(variable_lon, variable_lat, SCAN_ALT_TIE_POINTS, TIE_POINTS_FACTOR)

        assert np.allclose(return_lon, np.ones((10, 100)))
        assert return_lon.attrs == {"key_1": "value_lon_1", "key_2": "value_lon_2"}
        assert return_lon.name == "test_lon"
        assert return_lon.dims == ("num_pixels", "num_lines")

        assert np.allclose(return_lat, 6 * np.ones((10, 100)))
        assert return_lat.attrs == {"key_1": "value_lat_1", "key_2": "value_lat_2"}
        assert return_lat.name == "test_lat"
        assert return_lat.dims == ("num_pixels", "num_lines")

    def test_standardize_dims(self):
        """Test the standardize dims function."""
        test_variable = xr.DataArray(
            dims=("num_pixels", "num_lines"),
            name="test_data",
            attrs={
                "key_1": "value_lat_1",
                "key_2": "value_lat_2"
            },
            data=np.ones((10, 100)) * 1.
        )
        out_variable = self.reader._standardize_dims(test_variable)
        assert np.allclose(out_variable.values, np.ones((100, 10)))
        assert out_variable.dims == ("y", "x")
        assert out_variable.attrs["key_1"] == "value_lat_1"

    @mock.patch("satpy.readers.core.vii_nc.ViiNCBaseFileHandler._perform_calibration")
    @mock.patch("satpy.readers.core.vii_nc.ViiNCBaseFileHandler._perform_interpolation")
    @mock.patch("satpy.readers.core.vii_nc.ViiNCBaseFileHandler._perform_orthorectification")
    def test_dataset(self, po_, pi_, pc_):
        """Test the execution of the get_dataset function."""
        # Checks the correct execution of the get_dataset function with a valid file_key
        variable = self.reader.get_dataset(None, {"file_key": "data/measurement_data/tpw",
                                                  "calibration": None})
        pc_.assert_not_called()
        pi_.assert_not_called()
        po_.assert_not_called()

        assert np.allclose(variable.values, np.ones((100, 10)))
        assert variable.dims == ("y", "x")
        assert variable.attrs["test_attr"] == "attr"
        assert variable.attrs["units"] is None

        # Checks the correct execution of the get_dataset function with a valid file_key
        # and required calibration and interpolation
        self.reader.get_dataset(None, {"file_key": "data/measurement_data/tpw",
                                       "calibration": "reflectance",
                                       "interpolate": True,
                                       "standard_name": "longitude"})
        pc_.assert_called()
        pi_.assert_called()
        po_.assert_not_called()

        # Checks the correct execution of the get_dataset function with a valid file_key
        # and required orthorectification
        self.reader.orthorect = True
        self.reader.get_dataset(None, {"file_key": "data/measurement_data/tpw",
                                       "calibration": None,
                                       "orthorect_data": "test_orthorect_data"})
        po_.assert_called()

        # Checks the correct execution of the get_dataset function with an invalid file_key
        invalid_dataset = self.reader.get_dataset(None, {"file_key": "test_invalid", "calibration": None})
        # Checks that the function returns None
        assert invalid_dataset is None

        pc_.reset_mock()
        pi_.reset_mock()
        po_.reset_mock()

        # Checks the correct execution of the get_dataset function with a 'cached_longitude' file_key
        longitude = self.reader.get_dataset(None, {"file_key": "cached_longitude",
                                                   "calibration": "reflectance",
                                                   "interpolate": True})
        pc_.assert_not_called()
        pi_.assert_not_called()
        assert longitude[0, 0] == 1.0

        # Checks the correct execution of the get_dataset function with a 'cached_latitude' file_key
        latitude = self.reader.get_dataset(None, {"file_key": "cached_latitude",
                                                  "calibration": None})
        assert latitude[0, 0] == 2.0

        # Repeats some check with the reader where orthorectification and interpolation are inhibited
        # by means of the filetype_info flags

        pc_.reset_mock()
        pi_.reset_mock()
        po_.reset_mock()

        # Checks the correct execution of the get_dataset function with a valid file_key
        # and required calibration and interpolation
        self.reader_2.get_dataset(None, {"file_key": "data/measurement_data/tpw",
                                         "calibration": "reflectance",
                                         "interpolate": True,
                                         "standard_name": "longitude"})
        pc_.assert_called()
        pi_.assert_not_called()
        po_.assert_not_called()

        # Checks the correct execution of the get_dataset function with a valid file_key
        # and required orthorectification
        self.reader_2.get_dataset(None, {"file_key": "data/measurement_data/tpw",
                                         "calibration": None,
                                         "orthorect_data": "test_orthorect_data"})
        po_.assert_not_called()

        # Checks the correct execution of the get_dataset function with a 'cached_longitude' file_key
        longitude = self.reader_2.get_dataset(None, {"file_key": "cached_longitude",
                                                     "calibration": None})
        assert longitude[0, 0] == 100.0

        # Checks the correct execution of the get_dataset function with a 'cached_longitude' file_key
        # in a reader without defined longitude
        longitude = self.reader_3.get_dataset(None, {"file_key": "cached_longitude",
                                                     "calibration": "reflectance",
                                                     "interpolate": True})
        # Checks that the function returns None
        assert longitude is None



"""Tests for the Himawari L2 netCDF reader."""

import datetime as dt
import os

import numpy as np
import pytest
import xarray as xr

from satpy.readers.ahi_l2_nc import HIML2NCFileHandler
from satpy.tests.utils import make_dataid

rng = np.random.default_rng()
clmk_data = rng.integers(0, 3, (5500, 5500), dtype=np.uint16)
cprob_data = rng.uniform(0, 1, (5500, 5500))
lat_data = rng.uniform(-90, 90, (5500, 5500))
lon_data = rng.uniform(-180, 180, (5500, 5500))

start_time = dt.datetime(2023, 8, 24, 5, 40, 21)
end_time = dt.datetime(2023, 8, 24, 5, 49, 40)

dimensions = {"Columns": 5500, "Rows": 5500}

exp_ext = (-5499999.9012, -5499999.9012, 5499999.9012, 5499999.9012)

global_attrs = {"time_coverage_start": start_time.strftime("%Y-%m-%dT%H:%M:%SZ"),
                "time_coverage_end": end_time.strftime("%Y-%m-%dT%H:%M:%SZ"),
                "instrument_name": "AHI",
                "satellite_name": "Himawari-9",
                "cdm_data_type": "Full Disk",
                }

badarea_attrs = global_attrs.copy()
badarea_attrs["cdm_data_type"] = "bad_area"


def ahil2_filehandler(fname, platform="h09"):
    """Instantiate a Filehandler."""
    fileinfo = {"platform": platform}
    filetype = None
    fh = HIML2NCFileHandler(fname, fileinfo, filetype)
    return fh


@pytest.fixture(scope="module")
def himl2_filename(tmp_path_factory):
    """Create a fake himawari l2 file."""
    fname = f'{tmp_path_factory.mktemp("data")}/AHI-CMSK_v1r1_h09_s202308240540213_e202308240549407_c202308240557548.nc'
    ds = xr.Dataset({"CloudMask": (["Rows", "Columns"], clmk_data)},
                    coords={"Latitude": (["Rows", "Columns"], lat_data),
                            "Longitude": (["Rows", "Columns"], lon_data)},
                    attrs=global_attrs)
    ds.to_netcdf(fname)
    yield fname
    os.unlink(fname)


@pytest.fixture(scope="module")
def himl2_filename_bad(tmp_path_factory):
    """Create a fake himawari l2 file."""
    fname = f'{tmp_path_factory.mktemp("data")}/AHI-CMSK_v1r1_h09_s202308240540213_e202308240549407_c202308240557548.nc'
    ds = xr.Dataset({"CloudMask": (["Rows", "Columns"], clmk_data)},
                    coords={"Latitude": (["Rows", "Columns"], lat_data),
                            "Longitude": (["Rows", "Columns"], lon_data)},
                    attrs=badarea_attrs)
    ds.to_netcdf(fname)

    yield fname
    os.unlink(fname)


def test_startend(himl2_filename):
    """Test start and end times are set correctly."""
    fh = ahil2_filehandler(himl2_filename)
    assert fh.start_time == start_time
    assert fh.end_time == end_time


def test_ahi_l2_area_def(himl2_filename, caplog):
    """Test reader handles area definition correctly."""
    from pyproj import CRS

    ps = "+a=6378137 +h=35785863 +lon_0=140.7 +no_defs +proj=geos +rf=298.257024882273 +type=crs +units=m +x_0=0 +y_0=0"

    # Check case where input data is correct size.
    fh = ahil2_filehandler(himl2_filename)
    clmk_id = make_dataid(name="cloudmask")
    area_def = fh.get_area_def(clmk_id)
    assert area_def.width == dimensions["Columns"]
    assert area_def.height == dimensions["Rows"]
    assert np.allclose(area_def.area_extent, exp_ext)

    expected_crs = CRS(ps)
    assert area_def.crs == expected_crs

    # Check case where input data is incorrect size.
    fh = ahil2_filehandler(himl2_filename)
    fh.nlines = 3000
    with pytest.raises(ValueError, match="Input L2 file is not a full disk Himawari scene..*"):
        fh.get_area_def(clmk_id)


def test_bad_area_name(himl2_filename_bad):
    """Check case where area name is not correct."""
    global_attrs["cdm_data_type"] = "bad_area"
    with pytest.raises(ValueError, match="File is not a full disk scene"):
        ahil2_filehandler(himl2_filename_bad)
    global_attrs["cdm_data_type"] = "Full Disk"


def test_load_data(himl2_filename):
    """Test that data is loaded successfully."""
    fh = ahil2_filehandler(himl2_filename)
    clmk_id = make_dataid(name="cloudmask")
    clmk = fh.get_dataset(clmk_id, {"file_key": "CloudMask"})
    np.testing.assert_allclose(clmk.data, clmk_data)
    assert clmk.dtype == np.uint16
    assert clmk.attrs["sensor"] == "ahi"
    assert clmk.attrs["platform_name"] == "Himawari-9"
    assert clmk.attrs["platform_shortname"] == "h09"
    assert isinstance(clmk.attrs["start_time"], dt.datetime)
    assert isinstance(clmk.attrs["end_time"], dt.datetime)



"""Helpers for reading netcdf-based files."""

import logging
from contextlib import suppress

import dask.array as da
import netCDF4
import numpy as np
import xarray as xr

from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.readers.core.remote import open_file_or_filename
from satpy.readers.core.utils import np2str
from satpy.utils import get_legacy_chunk_size

LOG = logging.getLogger(__name__)
CHUNK_SIZE = get_legacy_chunk_size()


class NetCDF4FileHandler(BaseFileHandler):
    """Class for inspecting a NetCDF4 file and retrieving its metadata/header data.

    File information can be accessed using bracket notation. Variables are
    accessed by using:

        wrapper["var_name"]

    Or:

        wrapper["group/subgroup/var_name"]

    Attributes can be accessed by appending "/attr/attr_name" to the
    item string:

        wrapper["group/subgroup/var_name/attr/units"]

    Or for global attributes:

        wrapper["/attr/platform_short_name"]

    Or for all of global attributes:

        wrapper["/attrs"]

    Note that loading datasets requires reopening the original file
    (unless those datasets are cached, see below), but to get just the
    shape of the dataset append "/shape" to the item string:

        wrapper["group/subgroup/var_name/shape"]

    If your file has many small data variables that are frequently accessed,
    you may choose to cache some of them.  You can do this by passing a number,
    any variable smaller than this number in bytes will be read into RAM.
    Warning, this part of the API is provisional and subject to change.

    You may get an additional speedup by passing ``cache_handle=True``.  This
    will keep the netCDF4 dataset handles open throughout the lifetime of the
    object, and instead of using `xarray.open_dataset` to open every data
    variable, a dask array will be created "manually".  This may be useful if
    you have a dataset distributed over many files, such as for FCI.  Note
    that the coordinates will be missing in this case.  If you use this option,
    ``xarray_kwargs`` will have no effect.

    Args:
        filename (str): File to read
        filename_info (dict): Dictionary with filename information
        filetype_info (dict): Dictionary with filetype information
        auto_maskandscale (bool): Apply mask and scale factors
        xarray_kwargs (dict): Addition arguments to `xarray.open_dataset`
        cache_var_size (int): Cache variables smaller than this size.
        cache_handle (bool): Keep files open for lifetime of filehandler.
        engine (str or list of str): The engine to use for reading, either "netcdf4" or "h5netcdf". As a list, will try
            each engine until one works.

    """

    file_handle = None

    def __init__(self, filename, filename_info, filetype_info,
                 auto_maskandscale=False, xarray_kwargs=None,
                 cache_var_size=0, cache_handle=False, engine="netcdf4"):
        """Initialize object."""
        super().__init__(filename, filename_info, filetype_info)
        self.file_content = {}
        self.cached_file_content = {}
        self.engine = engine
        try:
            self.accessor, file_handle = self.get_accessor_and_filehandle()
        except IOError:
            LOG.exception(
                "Failed reading file %s. Possibly corrupted file", self.filename)
            raise
        self._set_file_handle_auto_maskandscale(file_handle, auto_maskandscale)
        self._set_xarray_kwargs(xarray_kwargs, auto_maskandscale)

        listed_variables = filetype_info.get("required_netcdf_variables")
        if listed_variables:
            self._collect_listed_variables(file_handle, listed_variables)
        else:
            self.collect_metadata("", file_handle)
            self.collect_dimensions("", file_handle)
        self.collect_cache_vars(cache_var_size)

        if cache_handle:
            self.file_handle = file_handle
        else:
            file_handle.close()

    def get_accessor_and_filehandle(self):
        """Choose the accessor based on the engine, and return in along with the file handle."""
        if not isinstance(self.engine, str):
            return get_accessor_and_filehandle_from_engines(self.filename, *self.engine)
        return get_accessor_and_filehandle_from_engine(self.filename, self.engine)

    @staticmethod
    def _set_file_handle_auto_maskandscale(file_handle, auto_maskandscale):
        if hasattr(file_handle, "set_auto_maskandscale"):
            file_handle.set_auto_maskandscale(auto_maskandscale)

    def _set_xarray_kwargs(self, xarray_kwargs, auto_maskandscale):
        self._xarray_kwargs = xarray_kwargs or {}
        self._xarray_kwargs.setdefault("chunks", CHUNK_SIZE)
        self._xarray_kwargs.setdefault("mask_and_scale", auto_maskandscale)

    def collect_metadata(self, name, obj):
        """Collect all file variables and attributes for the provided file object.

        This method also iterates through subgroups of the provided object.
        """
        # Look through each subgroup
        base_name = name + "/" if name else ""
        self._collect_groups_info(base_name, obj)
        self._collect_variables_info(base_name, obj)
        if not name:
            self._collect_global_attrs(obj)
        else:
            self._collect_attrs(name, obj)

    def _collect_groups_info(self, base_name, obj):
        for group_name, group_obj in obj.groups.items():
            full_group_name = base_name + group_name
            self.file_content[full_group_name] = group_obj
            self._collect_attrs(full_group_name, group_obj)
            self.collect_metadata(full_group_name, group_obj)

    def _collect_variables_info(self, base_name, obj):
        for var_name, var_obj in obj.variables.items():
            var_name = base_name + var_name
            self._collect_variable_info(var_name, var_obj)

    def _collect_variable_info(self, var_name, var_obj):
        self.file_content[var_name] = var_obj
        self.file_content[var_name + "/dtype"] = var_obj.dtype
        self.file_content[var_name + "/shape"] = var_obj.shape
        self.file_content[var_name + "/dimensions"] = var_obj.dimensions
        self._collect_attrs(var_name, var_obj)

    def _collect_listed_variables(self, file_handle, listed_variables):
        variable_name_replacements = self.filetype_info.get("variable_name_replacements")
        for itm in self._get_required_variable_names(listed_variables, variable_name_replacements):
            parts = itm.split("/")
            grp = file_handle
            for p in parts[:-1]:
                if p == "attr":
                    n = "/".join(parts)
                    self.file_content[n] = self._get_attr_value(grp, parts[-1])
                    break
                grp = grp[p]
            if p != "attr":
                var_obj = grp[parts[-1]]
                self._collect_variable_info(itm, var_obj)
                self.collect_dimensions(itm, grp)

    @staticmethod
    def _get_required_variable_names(listed_variables, variable_name_replacements):
        variable_names = []
        for var in listed_variables:
            if variable_name_replacements and "{" in var:
                _compose_replacement_names(variable_name_replacements, var, variable_names)
            else:
                variable_names.append(var)
        return variable_names

    def __del__(self):
        """Delete the file handler."""
        if self.file_handle is not None:
            with suppress(RuntimeError):
                self.file_handle.close()

    def _collect_global_attrs(self, obj):
        """Collect all the global attributes for the provided file object."""
        global_attrs = {}
        for key in self.accessor.get_object_attrs(obj):
            fc_key = f"/attr/{key}"
            value = self._get_attr_value(obj, key)
            self.file_content[fc_key] = global_attrs[key] = value
        self.file_content["/attrs"] = global_attrs

    def _collect_attrs(self, name, obj):
        """Collect all the attributes for the provided file object."""
        for key in self.accessor.get_object_attrs(obj):
            fc_key = f"{name}/attr/{key}"
            value = self._get_attr_value(obj, key)
            self.file_content[fc_key] = value

    def _get_attr_value(self, obj, key):
        value = self.accessor.get_attr(obj, key)
        try:
            value = np2str(value)
        except ValueError:
            pass
        return value

    def collect_dimensions(self, name, obj):
        """Collect dimensions."""
        for dim_name, dim_obj in obj.dimensions.items():
            dim_name = "{}/dimension/{}".format(name, dim_name)
            self.file_content[dim_name] = len(dim_obj)

    def collect_cache_vars(self, cache_var_size):
        """Collect data variables for caching.

        This method will collect some data variables and store them in RAM.
        This may be useful if some small variables are frequently accessed,
        to prevent needlessly frequently opening and closing the file, which
        in case of xarray is associated with some overhead.

        Should be called later than `collect_metadata`.

        Args:
            cache_var_size (int): Maximum size of the collected variables in bytes

        """
        if cache_var_size == 0:
            return

        cache_vars = self._collect_cache_var_names(cache_var_size)
        for var_name in cache_vars:
            v = self.file_content[var_name]
            arr = get_data_as_xarray(v)
            self.cached_file_content[var_name] = arr

    def _collect_cache_var_names(self, cache_var_size):
        return [varname for (varname, var)
                in self.file_content.items()
                if self.accessor.is_variable(var)
                and isinstance(var.dtype, np.dtype)  # vlen may be str
                and var.size * var.dtype.itemsize < cache_var_size]

    def __getitem__(self, key):
        """Get item for given key."""
        val = self.file_content[key]
        if self.accessor.is_variable(val):
            return self._get_variable(key, val)
        if self.accessor.is_group(val):
            return self._get_group(key, val)
        return val

    def _get_variable(self, key, val):
        """Get a variable from the netcdf file."""
        if key in self.cached_file_content:
            return self.cached_file_content[key]
        # these datasets are closed and inaccessible when the file is
        # closed, need to reopen
        # TODO: Handle HDF4 versus NetCDF3 versus NetCDF4
        parts = key.rsplit("/", 1)
        if len(parts) == 2:
            group, key = parts
        else:
            group = None
        if self.file_handle is not None:
            val = self._get_var_from_filehandle(group, key)
        else:
            val = self._get_var_from_xr(group, key)
        return val

    def _get_group(self, key, val):
        """Get a group from the netcdf file."""
        # Full groups are conveniently read with xr even if file_handle is available
        with xr.open_dataset(self.filename, group=key,
                             **self._xarray_kwargs) as nc:
            val = nc
        return val

    def _get_var_from_xr(self, group, key):
        with xr.open_dataset(self.filename, group=group,
                             **self._xarray_kwargs) as nc:
            val = nc[key]
            # Even though `chunks` is specified in the kwargs, xarray
            # uses dask.arrays only for data variables that have at least
            # one dimension; for zero-dimensional data variables (scalar),
            # it uses its own lazy loading for scalars.  When those are
            # accessed after file closure, xarray reopens the file without
            # closing it again.  This will leave potentially many open file
            # objects (which may in turn trigger a Segmentation Fault:
            # https://github.com/pydata/xarray/issues/2954#issuecomment-491221266
            if not val.chunks:
                val.load()
        return val

    def _get_var_from_filehandle(self, group, key):
        # Not getting coordinates as this is more work, therefore more
        # overhead, and those are not used downstream.
        if group is None:
            g = self.file_handle
        else:
            g = self.file_handle[group]
        v = g[key]
        attrs = self.accessor.get_object_attrs(v)
        x = xr.DataArray(
                da.from_array(v), dims=v.dimensions, attrs=attrs,
                name=v.name)
        return x

    def __contains__(self, item):
        """Get item from file content."""
        return item in self.file_content

    def get(self, item, default=None):
        """Get item."""
        if item in self:
            return self[item]
        else:
            return default

    def get_and_cache_npxr(self, var_name):
        """Get and cache variable as DataArray[numpy]."""
        if var_name in self.cached_file_content:
            return self.cached_file_content[var_name]
        v = self.file_content[var_name]
        if isinstance(v, xr.DataArray):
            val = v
        else:
            try:
                val = v[:]
                val = xr.DataArray(val, dims=v.dimensions,
                                   attrs=self.accessor.get_object_attrs(v),
                                   name=v.name)
            except IndexError:
                # Handle scalars
                val = v.__array__().item()
                val = xr.DataArray(val, dims=(), attrs={}, name=var_name)
            except AttributeError:
                # Handle strings
                val = v
        self.cached_file_content[var_name] = val
        return self.cached_file_content[var_name]

    def _get_attr(self, obj, key):
        return self.accessor.get_attr(obj, key)

    def _get_object_attrs(self, obj):
        return self.accessor.get_object_attrs(obj)

def _compose_replacement_names(variable_name_replacements, var, variable_names):
    for key in variable_name_replacements:
        vals = variable_name_replacements[key]
        for val in vals:
            if key in var:
                variable_names.append(var.format(**{key: val}))


def get_data_as_xarray(variable):
    """Get data in variable as xr.DataArray."""
    try:
        attrs = variable.attrs
    except AttributeError:
        # netCDF4 backend requires usage of __dict__ to get the attributes
        attrs = variable.__dict__
    try:
        data = variable[:]
    except (ValueError, IndexError):
        # Handle scalars for h5netcdf backend
        data = variable.__array__()

    arr = xr.DataArray(data, dims=variable.dimensions, attrs=attrs, name=variable.name)

    return arr


def choose_accessor_from_engine(engine):
    """Choose an accessor from engine."""
    if engine == "netcdf4":
        return NetCDF4Accessor()
    elif engine == "h5netcdf":
        return H5NetcdfAccessor()
    raise NotImplementedError(f"Engine {engine} not implemented.")


def get_accessor_and_filehandle_from_engine(filename, engine):
    """Choose an accessor from engine, and return in along with the file handle."""
    accessor = choose_accessor_from_engine(engine)
    file_handle = accessor.create_file_handle(filename)
    return accessor, file_handle


def get_accessor_and_filehandle_from_engines(filename, *engines):
    """Choose an accessor from the first possible engine, and return in along with the file handle."""
    for engine in engines:
        try:
            return get_accessor_and_filehandle_from_engine(filename, engine)
        except Exception as err:
            LOG.warning(str(err))
            continue
    else:
        raise RuntimeError("Could not work out an appropriate engine to open netCDF4 files")


class NetCDF4Accessor:
    """Accessor using the netCDF4 library as engine."""
    engine = "netcdf4"

    def create_file_handle(self, filename):
        """Create a file handle."""
        return netCDF4.Dataset(filename, "r")

    @staticmethod
    def is_variable(obj):
        """Check if obj is a variable."""
        return isinstance(obj, netCDF4.Variable)

    @staticmethod
    def is_group(obj):
        """Check if obj is a group."""
        return isinstance(obj, netCDF4.Group)

    @staticmethod
    def get_attr(obj, key):
        """Get an attribute from obj."""
        return getattr(obj, key)

    @staticmethod
    def get_object_attrs(obj):
        """Get the attributes for obj."""
        try:
            return obj.__dict__
        except KeyError:
            # Maybe unrecognised datatype, retrieve recoverable attributes.
            atts = {}
            for attname in obj.ncattrs():
                try:
                    atts[attname] = obj.getncattr(attname)
                except KeyError:
                    LOG.warning(f"Warning: Cannot load object ({obj.name}) attribute ({attname}).")
            return atts


class H5NetcdfAccessor:
    """Accessor using the h5netcdf library as engine."""
    engine = "h5netcdf"

    def create_file_handle(self, filename):
        """Create a file handle."""
        import h5netcdf
        f_obj = open_file_or_filename(filename)
        return h5netcdf.File(f_obj, "r")

    @staticmethod
    def is_variable(obj):
        """Check if obj is a variable."""
        import h5netcdf
        return isinstance(obj, h5netcdf.Variable)

    @staticmethod
    def is_group(obj):
        """Check if obj is a group."""
        import h5netcdf
        return isinstance(obj, h5netcdf.Group)

    @staticmethod
    def get_object_attrs(obj):
        """Get the attributes for obj."""
        return obj.attrs

    @staticmethod
    def get_attr(obj, key):
        """Get an attribute from obj."""
        return obj.attrs[key]


class NetCDF4FsspecFileHandler(NetCDF4FileHandler):
    """NetCDF4FileHandler implementation that allows accessing files on remote filesystems by switching engines."""

    def __init__(self, *args, engine=["netcdf4", "h5netcdf"], **kwargs):
        """Set up the instance with h5netcdf if netcdf4 does not work."""
        super().__init__(*args, engine=engine, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2022 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Utilities for reader tests."""

import inspect


def default_attr_processor(root, attr):
    """Do not change the attribute."""
    return attr


def fill_h5(root, contents, attr_processor=default_attr_processor):
    """Fill hdf5 file with the given contents.

    Args:
        root: hdf5 file rott
        contents: Contents to be written into the file
        attr_processor: A method for modifying attributes before they are
          written to the file.
    """
    for key, val in contents.items():
        if key in ["value", "attrs"]:
            continue
        if "value" in val:
            root[key] = val["value"]
        else:
            grp = root.create_group(key)
            fill_h5(grp, contents[key])
        if "attrs" in val:
            for attr_name, attr_val in val["attrs"].items():
                root[key].attrs[attr_name] = attr_processor(root, attr_val)


def get_jit_methods(module):
    """Get all jit-compiled methods in a module."""
    res = {}
    module_name = module.__name__
    members = inspect.getmembers(module)
    for member_name, obj in members:
        if _is_jit_method(obj):
            full_name = f"{module_name}.{member_name}"
            res[full_name] = obj
    return res


def _is_jit_method(obj):
    return hasattr(obj, "py_func")
